# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# ----------------------------------------------------------------------------

# Standard library imports
import os
import os.path as osp
from pathlib import Path
import sys
import threading
import traceback
from unittest.mock import Mock

# Third-party imports
import psutil
from pygments.token import Name
import pytest
from qtpy.QtWidgets import QMainWindow

# Local imports
from spyder.api.plugins import Plugins
from spyder.app.cli_options import get_options
from spyder.config.manager import CONF
from spyder.plugins.debugger.plugin import Debugger
from spyder.plugins.help.utils.sphinxify import CSS_PATH
from spyder.plugins.ipythonconsole.plugin import IPythonConsole
from spyder.plugins.ipythonconsole.utils.style import create_style_class
from spyder.utils.conda import get_list_conda_envs


# =============================================================================
# ---- Constants
# =============================================================================
SHELL_TIMEOUT = 40000 if os.name == 'nt' else 20000
NEW_DIR = 'new_workingdir'
PY312_OR_GREATER = sys.version_info[:2] >= (3, 12)


# =============================================================================
# ---- Pytest adjustments
# =============================================================================
@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"
    setattr(item, "rep_" + rep.when, rep)


# =============================================================================
# ---- Utillity Functions
# =============================================================================
def get_console_font_color(syntax_style):
    styles = create_style_class(syntax_style).styles
    font_color = styles[Name]
    return font_color


def get_console_background_color(style_sheet):
    background_color = style_sheet.split('background-color:')[1]
    background_color = background_color.split(';')[0]
    return background_color


def get_conda_test_env():
    """
    Return the full prefix path of the env used to test kernel activation and
    its executable.
    """
    # Get conda env to use
    test_env_executable = get_list_conda_envs()['Conda: spytest-ž'][0]

    # Get the env prefix
    if os.name == 'nt':
        test_env_prefix = osp.dirname(test_env_executable)
    else:
        test_env_prefix = osp.dirname(osp.dirname(test_env_executable))

    return (test_env_prefix, test_env_executable)


# =============================================================================
# ---- Fixtures
# =============================================================================
@pytest.fixture
def ipyconsole(qtbot, request, tmpdir):
    """IPython console fixture."""
    configuration = CONF
    no_web_widgets = request.node.get_closest_marker('no_web_widgets')

    class MainWindowMock(QMainWindow):

        def __init__(self):
            # This avoids using the cli options passed to pytest
            sys_argv = [sys.argv[0]]
            self._cli_options = get_options(sys_argv)[0]
            if no_web_widgets:
                self._cli_options.no_web_widgets = True
            super().__init__()

        def __getattr__(self, attr):
            if attr == 'consoles_menu_actions':
                return []
            elif attr == 'editor':
                return None
            else:
                return Mock()

    # Tests assume inline backend
    configuration.set('ipython_console', 'pylab/backend', 'inline')

    # Start the console in a fixed working directory
    use_startup_wdir = request.node.get_closest_marker('use_startup_wdir')
    if use_startup_wdir:
        new_wdir = str(tmpdir.mkdir(NEW_DIR))
        configuration.set(
            'workingdir',
            'startup/use_project_or_home_directory',
            False
        )
        configuration.set('workingdir', 'startup/use_fixed_directory', True)
        configuration.set('workingdir', 'startup/fixed_directory', new_wdir)
    else:
        configuration.set(
            'workingdir',
            'startup/use_project_or_home_directory',
            True
        )
        configuration.set('workingdir', 'startup/use_fixed_directory', False)

    # Use the automatic backend if requested
    auto_backend = request.node.get_closest_marker('auto_backend')
    if auto_backend:
        configuration.set('ipython_console', 'pylab/backend', 'auto')

    # Use the Tkinter backend if requested
    tk_backend = request.node.get_closest_marker('tk_backend')
    if tk_backend:
        configuration.set('ipython_console', 'pylab/backend', 'tk')

    # Start a Pylab client if requested
    pylab_client = request.node.get_closest_marker('pylab_client')
    special = "pylab" if pylab_client else None

    # Start a Sympy client if requested
    sympy_client = request.node.get_closest_marker('sympy_client')
    special = "sympy" if sympy_client else special

    # Start a Cython client if requested
    cython_client = request.node.get_closest_marker('cython_client')
    special = "cython" if cython_client else special

    # Start a specific env client if requested
    environment_client = request.node.get_closest_marker(
        'environment_client')
    given_name = None
    path_to_custom_interpreter = None
    if environment_client:
        given_name = 'spytest-ž'
        path_to_custom_interpreter = get_conda_test_env()[1]

    # Use an external interpreter if requested
    external_interpreter = request.node.get_closest_marker(
        'external_interpreter')
    if external_interpreter:
        configuration.set('main_interpreter', 'default', False)
        configuration.set('main_interpreter', 'executable', sys.executable)
    else:
        configuration.set('main_interpreter', 'default', True)
        configuration.set('main_interpreter', 'executable', '')

    # Use the test environment interpreter if requested
    test_environment_interpreter = request.node.get_closest_marker(
        'test_environment_interpreter')
    if test_environment_interpreter:
        configuration.set('main_interpreter', 'default', False)
        configuration.set(
            'main_interpreter', 'executable', get_conda_test_env()[1])
    else:
        configuration.set('main_interpreter', 'default', True)
        configuration.set('main_interpreter', 'executable', '')

    # Conf css_path in the Appeareance plugin
    configuration.set('appearance', 'css_path', CSS_PATH)

    # Create the console and a new client and set environment
    os.environ['IPYCONSOLE_TESTING'] = 'True'
    window = MainWindowMock()
    console = IPythonConsole(parent=window, configuration=configuration)

    # connect to a debugger plugin
    debugger = Debugger(parent=window, configuration=configuration)

    def get_plugin(name):
        if name == Plugins.IPythonConsole:
            return console
        return None

    debugger.get_plugin = get_plugin
    debugger.on_ipython_console_available()

    # Plugin setup
    console.on_initialize()
    console._register()
    console.get_widget().matplotlib_status.register_ipythonconsole(console)

    # Register handlers to run cells.
    def get_file_code(fname, save_all=True):
        """
        Get code from a file.

        save_all is necessary to keep consistency with the handler registered
        in the editor.
        """
        path = Path(fname)
        return path.read_text()

    def get_cell(cell_id, fname):
        """
        Get cell code from a file.

        For now this only works with unnamed cells and cell separators of the
        form `# %%`.
        """
        path = Path(fname)
        contents = path.read_text()
        cells = contents.split("# %%")
        return cells[int(cell_id)]

    console.register_spyder_kernel_call_handler('get_file_code', get_file_code)
    console.register_spyder_kernel_call_handler('run_cell', get_cell)

    # Start client and show window
    console.create_new_client(
        special=special,
        given_name=given_name,
        path_to_custom_interpreter=path_to_custom_interpreter
    )
    window.setCentralWidget(console.get_widget())

    if os.name == 'nt':
        qtbot.addWidget(window)

    with qtbot.waitExposed(window):
        window.resize(640, 480)
        window.show()

    # Set exclamation mark to True
    configuration.set('debugger', 'pdb_use_exclamation_mark', True)

    # Create new client for Matplotlb backend tests
    if auto_backend or tk_backend:
        qtbot.wait(SHELL_TIMEOUT)
        console.create_new_client()

    # Wait until the window is fully up
    qtbot.waitUntil(lambda: console.get_current_shellwidget() is not None)
    shell = console.get_current_shellwidget()
    try:
        if test_environment_interpreter:
            # conda version is not always up to date, so a version warning
            # might be displayed, so shell.spyder_kernel_ready will not be True
            qtbot.waitUntil(
                lambda: shell._prompt_html is not None,
                timeout=SHELL_TIMEOUT)
        else:
            qtbot.waitUntil(
                lambda: (
                    shell.spyder_kernel_ready
                    and shell._prompt_html is not None
                ),
                timeout=SHELL_TIMEOUT)
    except Exception:
        # Print content of shellwidget and close window
        print(console.get_current_shellwidget()._control.toPlainText())
        client = console.get_current_client()
        if client.info_page != client.blank_page:
            print('info_page')
            print(client.info_page)
        raise

    # Check for thread or open file leaks
    known_leak = request.node.get_closest_marker('known_leak')

    if os.name != 'nt' and not known_leak:
        # _DummyThread are created if current_thread() is called from them.
        # They will always leak (From python doc) so we ignore them.
        init_threads = [
            repr(thread) for thread in threading.enumerate()
            if not isinstance(thread, threading._DummyThread)]
        proc = psutil.Process()
        init_files = [repr(f) for f in proc.open_files()]
        init_subprocesses = [repr(f) for f in proc.children()]

    yield console

    # Print shell content if failed
    if request.node.rep_setup.passed:
        if request.node.rep_call.failed:
            # Print content of shellwidget and close window
            print(console.get_current_shellwidget()._control.toPlainText())
            client = console.get_current_client()
            if client.info_page != client.blank_page:
                print('info_page')
                print(client.info_page)

    # Close
    console.on_close()
    os.environ.pop('IPYCONSOLE_TESTING')

    if os.name == 'nt' or known_leak:
        # Do not test for leaks
        return

    def show_diff(init_list, now_list, name):
        sys.stderr.write(f"Extra {name} before test:\n")
        for item in init_list:
            if item in now_list:
                now_list.remove(item)
            else:
                sys.stderr.write(item + "\n")
        sys.stderr.write(f"Extra {name} after test:\n")
        for item in now_list:
            sys.stderr.write(item + "\n")

    # The test is not allowed to open new files or threads.
    try:
        def threads_condition():
            threads = [
                thread for thread in threading.enumerate()
                if not isinstance(thread, threading._DummyThread)]
            return (len(init_threads) >= len(threads))

        qtbot.waitUntil(threads_condition, timeout=SHELL_TIMEOUT)
    except Exception:
        now_threads = [
            thread for thread in threading.enumerate()
            if not isinstance(thread, threading._DummyThread)]
        threads = [repr(t) for t in now_threads]
        show_diff(init_threads, threads, "thread")
        sys.stderr.write("Running Threads stacks:\n")
        now_thread_ids = [t.ident for t in now_threads]
        for thread_id, frame in sys._current_frames().items():
            if thread_id in now_thread_ids:
                sys.stderr.write("\nThread " + str(threads) + ":\n")
                traceback.print_stack(frame)
        raise

    try:
        # -1 from closed client
        qtbot.waitUntil(lambda: (
            len(init_subprocesses) - 1 >= len(proc.children())),
            timeout=SHELL_TIMEOUT)
    except Exception:
        subprocesses = [repr(f) for f in proc.children()]
        show_diff(init_subprocesses, subprocesses, "processes")
        raise

    try:
        qtbot.waitUntil(
            lambda: (len(init_files) >= len(proc.open_files())),
            timeout=SHELL_TIMEOUT)
    except Exception:
        files = [repr(f) for f in proc.open_files()]
        show_diff(init_files, files, "files")
        raise


@pytest.fixture
def mpl_rc_file(tmp_path):
    """Create matplotlibrc file"""
    file_contents = """
figure.dpi: 99
figure.figsize: 9, 9
figure.subplot.bottom: 0.9
font.size: 9
"""
    rc_file = str(tmp_path / 'matplotlibrc')
    with open(rc_file, 'w') as f:
        f.write(file_contents)
    os.environ['MATPLOTLIBRC'] = rc_file

    yield

    os.environ.pop('MATPLOTLIBRC')
    os.remove(rc_file)



# -*- coding: utf-8 -*-

# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

import json
import os.path as osp

import pytest
from diff_match_patch import diff_match_patch
from spyder.plugins.completion.api import CompletionRequestTypes
from spyder.plugins.completion.providers.fallback.utils import get_words


DATA_PATH = osp.join(osp.dirname(osp.abspath(__file__)), "data")
TOKENS_FILE = osp.join(DATA_PATH, 'tokens.json')

language_list = ['c', 'cpp', 'css', 'erl', 'ex', 'html', 'java', 'jl',
                 'md', 'py', 'R']
extension_map = {
    'c': 'C', 'cpp': 'C++', 'css': 'CSS', 'erl': 'Erlang', 'ex': "Elixir",
    'html': 'HTML', 'java': 'Java', 'jl': 'Julia', 'md': 'Markdown',
    'py': 'Python', 'R': 'R'
}

TEST_FILE = """
# This is a test file
a = 2
"""

TEST_FILE_UPDATE = """
# This is a test file
a = 2

def func(args):
    pass
"""


@pytest.fixture
def tokens_fixture():
    with open(TOKENS_FILE, 'r') as f:
        tokens = json.load(f)
    return tokens


@pytest.fixture
def file_fixture(tokens_fixture, request):
    ext = request.param
    filename = 'example.{0}'.format(ext)
    example_file = osp.join(DATA_PATH, filename)
    with open(example_file) as f:
        contents = f.read()
    return filename, set(tokens_fixture[filename]), contents


@pytest.fixture(scope="module")
def fallback_fixture(fallback_completions, qtbot_module, request):
    fallback, completions = fallback_completions
    diff_match = diff_match_patch()
    return fallback, completions, diff_match


def test_file_open_close(qtbot_module, fallback_fixture):
    fallback, completions, diff_match = fallback_fixture

    open_request = {
        'file': 'test.py',
        'text': TEST_FILE,
        'offset': -1,
    }
    fallback.send_request(
        'python', CompletionRequestTypes.DOCUMENT_DID_OPEN, open_request)
    qtbot_module.wait(1000)
    assert 'test.py' in fallback.fallback_actor.file_tokens

    close_request = {
        'file': 'test.py',
    }
    fallback.send_request(
        'python', CompletionRequestTypes.DOCUMENT_DID_CLOSE, close_request)
    qtbot_module.wait(1000)
    assert 'test.py' not in fallback.fallback_actor.file_tokens


def test_get_words():
    source = 'foo bar123 baz car456'
    tokens = get_words(source, 5, 'python')
    assert set(tokens) == {'foo', 'baz', 'car456'}


@pytest.mark.parametrize('file_fixture', language_list, indirect=True)
def test_tokenize(qtbot_module, fallback_fixture, file_fixture):
    filename, expected_tokens, contents = file_fixture
    _, ext = osp.splitext(filename)
    language = extension_map[ext[1:]]
    fallback, completions, diff_match = fallback_fixture
    # diff = diff_match.patch_make('', contents)
    open_request = {
        'file': filename,
        'text': contents,
        'offset': len(contents),
    }
    fallback.send_request(
        language, CompletionRequestTypes.DOCUMENT_DID_OPEN, open_request)
    qtbot_module.wait(1000)

    tokens_request = {
        'file': filename,
        'current_word': ''
    }
    with qtbot_module.waitSignal(completions.sig_recv_tokens,
                                 timeout=3000) as blocker:
        fallback.send_request(
            language,
            CompletionRequestTypes.DOCUMENT_COMPLETION,
            tokens_request
        )
    tokens = blocker.args
    tokens = {token['insertText'] for token in tokens[0]}
    assert len(expected_tokens - tokens) == 0


def test_token_update(qtbot_module, fallback_fixture):
    fallback, completions, diff_match = fallback_fixture

    # diff = diff_match.patch_make('', TEST_FILE)
    open_request = {
        'file': 'test.py',
        'text': TEST_FILE,
        'offset': len(TEST_FILE),
    }
    fallback.send_request(
        'python', CompletionRequestTypes.DOCUMENT_DID_OPEN, open_request)
    qtbot_module.wait(1000)

    tokens_request = {
        'file': 'test.py',
        'current_word': ''
    }
    with qtbot_module.waitSignal(completions.sig_recv_tokens,
                                 timeout=3000) as blocker:
        fallback.send_request(
            'python',
            CompletionRequestTypes.DOCUMENT_COMPLETION,
            tokens_request
        )
    initial_tokens = blocker.args[0]
    initial_tokens = {token['insertText'] for token in initial_tokens}
    assert 'args' not in initial_tokens

    diff = diff_match.patch_make(TEST_FILE, TEST_FILE_UPDATE)
    update_request = {
        'file': 'test.py',
        'diff': diff,
        'offset': len(diff),
    }
    fallback.send_request(
        'python', CompletionRequestTypes.DOCUMENT_DID_CHANGE, update_request)
    qtbot_module.wait(1000)
    with qtbot_module.waitSignal(completions.sig_recv_tokens,
                                 timeout=3000) as blocker:
        fallback.send_request(
            'python',
            CompletionRequestTypes.DOCUMENT_COMPLETION,
            tokens_request
        )
    updated_tokens = blocker.args[0]
    updated_tokens = {token['insertText'] for token in updated_tokens}
    assert 'args' in updated_tokens



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tests for the dataframe editor.
"""

# Standard library imports
import os
import sys
from datetime import datetime
from unittest.mock import Mock, patch, ANY

# Third party imports
from flaky import flaky
import numpy
from packaging.version import parse
from pandas import (
    __version__ as pandas_version, DataFrame, date_range, read_csv, concat,
    Index, RangeIndex, MultiIndex, CategoricalIndex, Series)
from pandas.testing import assert_frame_equal
import pytest
from qtpy.QtGui import QColor
from qtpy.QtCore import Qt, QTimer
from qtpy.QtWidgets import QDialog, QInputDialog, QMessageBox

# Local imports
from spyder.utils.programs import is_module_installed
from spyder.utils.test import close_message_box
from spyder.plugins.variableexplorer.widgets import dataframeeditor
from spyder.plugins.variableexplorer.widgets.dataframeeditor import (
    DataFrameEditor, DataFrameModel, COLS_TO_LOAD, LARGE_COLS)


# =============================================================================
# Constants
# =============================================================================
FILES_PATH = os.path.dirname(os.path.realpath(__file__))


# =============================================================================
# Utility functions
# =============================================================================
def colorclose(color, hsva_expected):
    """
    Compares HSV values which are stored as 16-bit integers.
    """
    hsva_actual = color.getHsvF()
    return all(abs(a-b) <= 2**(-16)
               for (a, b) in zip(hsva_actual, hsva_expected))


def data(dfm, i, j):
    return dfm.data(dfm.createIndex(i, j))


def bgcolor(dfm, i, j):
    return dfm.get_bgcolor(dfm.createIndex(i, j))


def data_header(dfh, i, j, role=Qt.DisplayRole):
    return dfh.data(dfh.createIndex(i, j), role)


def data_index(dfi, i, j, role=Qt.DisplayRole):
    return dfi.data(dfi.createIndex(i, j), role)


def generate_pandas_indexes():
    """Creates a dictionary of many possible pandas indexes."""
    # Float64Index was removed in Pandas 2.0
    if parse(pandas_version) >= parse('2.0.0'):
        float_index = 'Index'
    else:
        float_index = 'Float64Index'

    indexes = {
        'RangeIndex': RangeIndex(0, 20),
        float_index: Index([i/10 for i in range(20)]),
        'DatetimeIndex': date_range(start='2017-01-01', periods=20, freq='D'),
        'MultiIndex': MultiIndex.from_product(
            [list('ABCDEFGHIJ'), ('foo', 'bar')], names=['first', 'second']),
        'CategoricalIndex': CategoricalIndex(list('abcaadaccbbacabacccb'),
                                             categories=['a', 'b', 'c']),
    }

    if parse(pandas_version) < parse('2.0.0'):
        indexes['Index'] = Index(list('ABCDEFGHIJKLMNOPQRST'))

    return indexes


# =============================================================================
# Tests
# =============================================================================
def test_dataframemodel_index_sort(qtbot):
    """Validate the data in the model for index when sorting."""
    ds = Series(numpy.arange(10))
    editor = DataFrameEditor(None)
    editor.setup_and_check(ds)
    index = editor.table_index.model()

    index.sort(-1, order=Qt.AscendingOrder)
    assert data_index(index, 0, 0, Qt.DisplayRole) == '0'
    assert data_index(index, 9, 0, Qt.DisplayRole) == '9'

    index.sort(-1, order=Qt.DescendingOrder)
    assert data_index(index, 0, 0, Qt.DisplayRole) == '9'
    assert data_index(index, 9, 0, Qt.DisplayRole) == '0'


def test_dataframe_to_type(qtbot):
    """Regression test for spyder-ide/spyder#12296"""
    # Setup editor
    d = {'col1': [1, 2], 'col2': [3, 4]}
    df = DataFrame(data=d)
    editor = DataFrameEditor()
    assert editor.setup_and_check(df, 'Test DataFrame To action')
    with qtbot.waitExposed(editor):
        editor.show()

    # Check editor doesn't have changes to save and select an initial element
    assert not editor.btn_save_and_close.isEnabled()
    view = editor.dataTable
    view.setCurrentIndex(view.model().index(0, 0))

    # Show context menu, go down until `Convert to`, and open submenu
    view.menu.render()
    view.menu.show()
    for _ in range(100):
        activeAction = view.menu.activeAction()
        if activeAction and activeAction.text() == 'Convert to':
            qtbot.keyPress(view.menu, Qt.Key_Return)
            break
        qtbot.keyPress(view.menu, Qt.Key_Down)
    else:
        raise RuntimeError('Item "Convert to" not found')

    # Select first option, which is `To bool`
    submenu = activeAction.menu()
    qtbot.keyPress(submenu, Qt.Key_Return)
    qtbot.wait(1000)

    # Check that changes where made from the editor
    assert editor.btn_save_and_close.isEnabled()


def test_dataframe_editor_shows_scrollbar(qtbot):
    """
    Test the dataframe editor shows a scrollbar when opening a large dataframe.
    Regression test for spyder-ide/spyder#21627 .
    """
    df = DataFrame(numpy.zeros((100, 100)))
    editor = DataFrameEditor()
    editor.setup_and_check(df)
    with qtbot.waitExposed(editor):
        editor.show()

    assert editor.dataTable.horizontalScrollBar().isVisible()


def test_dataframe_editor_scroll(qtbot):
    """
    Test that when opening a "large" dataframe, only a part of it is initially
    loaded in the editor window. When scrolling past that part, the rest is
    loaded. When moving to the right-most column and sorting it, the view
    stay scrolled to the right end.

    Regression test for spyder-ide/spyder#21627 .
    """

    # Make DataFrame with LARGE_COLS + 5 columns
    df = DataFrame(numpy.zeros((10, LARGE_COLS + 5)))
    editor = DataFrameEditor()
    editor.setup_and_check(df)
    model = editor.dataModel
    with qtbot.waitExposed(editor):
        editor.show()

    # Check that initially, COLS_TO_LOAD columns are loaded in the editor
    assert model.rowCount() == 10
    assert model.columnCount() == COLS_TO_LOAD

    # Press the End key to move to the right and wait
    view = editor.dataTable
    view.setCurrentIndex(view.model().index(0, 0))
    qtbot.keyPress(view, Qt.Key_End)

    # Check that now all the columns are loaded in the editor
    def check_column_count():
        assert model.columnCount() == LARGE_COLS + 5

    qtbot.waitUntil(check_column_count)

    # Press the End key to move to the right and wait
    qtbot.keyPress(view, Qt.Key_End)
    scrollbar = editor.dataTable.horizontalScrollBar()

    # Check that we are at the far right
    def check_at_far_right():
        assert scrollbar.value() == scrollbar.maximum()

    qtbot.waitUntil(check_at_far_right)

    # Sort the rightmost column
    old_index_model = editor.table_index.model()
    view.sortByColumn(model.columnCount() - 1)

    # Wait until the model for the index is updated
    def check_index_model_updated():
        assert editor.table_index.model() != old_index_model

    qtbot.waitUntil(check_index_model_updated)

    # Check that we are at the far right
    assert scrollbar.value() == scrollbar.maximum()


def test_dataframe_datetimeindex(qtbot):
    """Regression test for spyder-ide/spyder#11129 ."""
    ds = Series(
        numpy.arange(10),
        index=date_range('2019-01-01', periods=10))
    editor = DataFrameEditor(None)
    editor.setup_and_check(ds)
    index = editor.table_index.model()
    assert data_index(index, 0, 0) == '2019-01-01 00:00:00'
    assert data_index(index, 9, 0) == '2019-01-10 00:00:00'


def test_dataframe_simpleindex(qtbot):
    """Test to validate proper creation and handling of a simpleindex."""
    df = DataFrame(numpy.random.randn(6, 6))
    editor = DataFrameEditor(None)
    editor.setup_and_check(df)
    header = editor.table_header.model()
    assert header.headerData(0, Qt.Horizontal,
                             Qt.DisplayRole) == "0"
    assert header.headerData(1, Qt.Horizontal,
                             Qt.DisplayRole) == "1"
    assert header.headerData(5, Qt.Horizontal,
                             Qt.DisplayRole) == "5"


def test_dataframe_simpleindex_custom_columns():
    """Test to validate proper creation and handling of custom simpleindex."""
    df = DataFrame(numpy.random.randn(10, 5),
                   columns=['a', 'b', 'c', 'd', 'e'])
    editor = DataFrameEditor(None)
    editor.setup_and_check(df)
    header = editor.table_header.model()
    assert header.headerData(0, Qt.Horizontal,
                             Qt.DisplayRole) == "a"
    assert header.headerData(1, Qt.Horizontal,
                             Qt.DisplayRole) == "b"
    assert header.headerData(4, Qt.Horizontal,
                             Qt.DisplayRole) == "e"


def test_dataframe_multiindex():
    """Test to validate proper creation and handling of a multiindex."""
    arrays = [numpy.array(['bar', 'bar', 'baz', 'baz',
                           'foo', 'foo', 'qux', 'qux']),
              numpy.array(['one', 'two', 'one', 'two',
                           'one', 'two', 'one', 'two'])]
    tuples = list(zip(*arrays))
    index = MultiIndex.from_tuples(tuples, names=['first', 'second'])
    df = DataFrame(numpy.random.randn(6, 6), index=index[:6],
                   columns=index[:6])
    editor = DataFrameEditor(None)
    editor.setup_and_check(df)
    header = editor.table_header.model()
    assert header.headerData(0, Qt.Horizontal,
                             Qt.DisplayRole) == 0
    assert data_header(header, 0, 0) == 'bar'
    assert data_header(header, 1, 0) == 'one'
    assert data_header(header, 0, 1) == 'bar'
    assert data_header(header, 1, 1) == 'two'
    assert data_header(header, 0, 2) == 'baz'
    assert data_header(header, 1, 2) == 'one'
    assert data_header(header, 0, 3) == 'baz'
    assert data_header(header, 1, 3) == 'two'
    assert data_header(header, 0, 4) == 'foo'
    assert data_header(header, 1, 4) == 'one'
    assert data_header(header, 0, 5) == 'foo'
    assert data_header(header, 1, 5) == 'two'


def test_header_bom():
    """Test for BOM data in the headers."""
    df = read_csv(os.path.join(FILES_PATH, 'issue_2514.csv'))
    editor = DataFrameEditor(None)
    editor.setup_and_check(df)
    header = editor.table_header.model()
    assert header.headerData(0, Qt.Horizontal,
                             Qt.DisplayRole) == "Date (MMM-YY)"


@pytest.mark.skipif(is_module_installed('pandas', '<0.19'),
                    reason="It doesn't work for Pandas 0.19-")
def test_header_encoding():
    """Test for header encoding handling."""
    df = read_csv(os.path.join(FILES_PATH, 'issue_3896.csv'))
    editor = DataFrameEditor(None)
    editor.setup_and_check(df)
    header = editor.table_header.model()
    assert header.headerData(0, Qt.Horizontal,
                             Qt.DisplayRole) == "Unnamed: 0"
    assert "Unieke_Idcode" in header.headerData(1, Qt.Horizontal,
                                                Qt.DisplayRole)
    assert header.headerData(2, Qt.Horizontal,
                             Qt.DisplayRole) == "a"
    assert header.headerData(3, Qt.Horizontal,
                             Qt.DisplayRole) == "b"
    assert header.headerData(4, Qt.Horizontal,
                             Qt.DisplayRole) == "c"
    assert header.headerData(5, Qt.Horizontal,
                             Qt.DisplayRole) == "d"


def test_dataframemodel_basic():
    df = DataFrame({'colA': [1, 3], 'colB': ['c', 'a']})
    dfm = DataFrameModel(df)
    assert dfm.rowCount() == 2
    assert dfm.columnCount() == 2
    assert data(dfm, 0, 0) == '1'
    assert data(dfm, 0, 1) == 'c'
    assert data(dfm, 1, 0) == '3'
    assert data(dfm, 1, 1) == 'a'


def test_dataframemodel_sort():
    """Validate the data in the model."""
    df = DataFrame({'colA': [1, 3], 'colB': ['c', 'a']})
    dfm = DataFrameModel(df)
    dfm.sort(1)
    assert data(dfm, 0, 0) == '3'
    assert data(dfm, 1, 0) == '1'
    assert data(dfm, 0, 1) == 'a'
    assert data(dfm, 1, 1) == 'c'


def test_dataframemodel_sort_is_stable():   # cf. spyder-ide/spyder#3010.
    """Validate the sort function."""
    df = DataFrame([[2, 14], [2, 13], [2, 16], [1, 3], [2, 9], [1, 15],
                    [1, 17], [2, 2], [2, 10], [1, 6], [2, 5], [2, 8],
                    [1, 11], [1, 1], [1, 12], [1, 4], [2, 7]])
    dfm = DataFrameModel(df)
    dfm.sort(1)
    dfm.sort(0)
    col2 = [data(dfm, i, 1) for i in range(len(df))]
    assert col2 == [str(x) for x in [1, 3, 4, 6, 11, 12, 15, 17,
                                     2, 5, 7, 8, 9, 10, 13, 14, 16]]


def test_dataframemodel_max_min_col_update():
    df = DataFrame([[1, 2.0], [2, 2.5], [3, 9.0]])
    dfm = DataFrameModel(df)
    assert dfm.max_min_col == [[3, 1], [9.0, 2.0]]


def test_dataframemodel_max_min_col_update_constant():
    df = DataFrame([[1, 2.0], [1, 2.0], [1, 2.0]])
    dfm = DataFrameModel(df)
    assert dfm.max_min_col == [[1, 0], [2.0, 1.0]]


def test_dataframemodel_with_timezone_aware_timestamps():
    # cf. spyder-ide/spyder#2940.
    df = DataFrame([x] for x in date_range('20150101', periods=5, tz='UTC'))
    dfm = DataFrameModel(df)
    assert dfm.max_min_col == [None]


def test_dataframemodel_with_categories():  # cf. spyder-ide/spyder#3308.
    df = DataFrame({"id": [1, 2, 3, 4, 5, 6],
                    "raw_grade": ['a', 'b', 'b', 'a', 'a', 'e']})
    df["grade"] = df["raw_grade"].astype("category")
    dfm = DataFrameModel(df)
    assert dfm.max_min_col == [[6, 1], None, None]


def test_dataframemodel_get_bgcolor_with_numbers():
    df = DataFrame([[0, 10], [1, 20], [2, 40]])
    dfm = DataFrameModel(df)
    h0 = dataframeeditor.BACKGROUND_NUMBER_MINHUE
    dh = dataframeeditor.BACKGROUND_NUMBER_HUERANGE
    s = dataframeeditor.BACKGROUND_NUMBER_SATURATION
    v = dataframeeditor.BACKGROUND_NUMBER_VALUE
    a = dataframeeditor.BACKGROUND_NUMBER_ALPHA
    assert colorclose(bgcolor(dfm, 0, 0), (h0 + dh,         s, v, a))
    assert colorclose(bgcolor(dfm, 1, 0), (h0 + 1 / 2 * dh, s, v, a))
    assert colorclose(bgcolor(dfm, 2, 0), (h0,              s, v, a))
    assert colorclose(bgcolor(dfm, 0, 1), (h0 + dh,         s, v, a))
    assert colorclose(bgcolor(dfm, 1, 1), (h0 + 2 / 3 * dh, s, v, a))
    assert colorclose(bgcolor(dfm, 2, 1), (h0,              s, v, a))


def test_dataframemodel_get_bgcolor_with_numbers_using_global_max():
    df = DataFrame([[0, 10], [1, 20], [2, 40]])
    dfm = DataFrameModel(df)
    dfm.colum_avg(0)
    h0 = dataframeeditor.BACKGROUND_NUMBER_MINHUE
    dh = dataframeeditor.BACKGROUND_NUMBER_HUERANGE
    s = dataframeeditor.BACKGROUND_NUMBER_SATURATION
    v = dataframeeditor.BACKGROUND_NUMBER_VALUE
    a = dataframeeditor.BACKGROUND_NUMBER_ALPHA
    assert colorclose(bgcolor(dfm, 0, 0), (h0 + dh,           s, v, a))
    assert colorclose(bgcolor(dfm, 1, 0), (h0 + 39 / 40 * dh, s, v, a))
    assert colorclose(bgcolor(dfm, 2, 0), (h0 + 38 / 40 * dh, s, v, a))
    assert colorclose(bgcolor(dfm, 0, 1), (h0 + 30 / 40 * dh, s, v, a))
    assert colorclose(bgcolor(dfm, 1, 1), (h0 + 20 / 40 * dh, s, v, a))
    assert colorclose(bgcolor(dfm, 2, 1), (h0,                s, v, a))


def test_dataframemodel_get_bgcolor_with_string():
    """Validate the color of the cell when a string is the data."""
    df = DataFrame([['xxx']])
    dfm = DataFrameModel(df)
    h, s, v, dummy = \
        QColor(dataframeeditor.BACKGROUND_NONNUMBER_COLOR).getHsvF()
    a = dataframeeditor.BACKGROUND_STRING_ALPHA
    assert colorclose(bgcolor(dfm, 0, 0), (h, s, v, a))


def test_dataframemodel_get_bgcolor_with_object():
    df = DataFrame([[None]])
    dfm = DataFrameModel(df)
    h, s, v, dummy = \
        QColor(dataframeeditor.BACKGROUND_NONNUMBER_COLOR).getHsvF()
    a = dataframeeditor.BACKGROUND_MISC_ALPHA
    assert colorclose(bgcolor(dfm, 0, 0), (h, s, v, a))


def test_dataframemodel_get_bgcolor_with_missings():
    """
    Test that df bg colors are correct for missing values of various types.

    The types `bool`, `object`, `datetime`, and `timedelta` are omitted,
    because missings have no different background there yet.
    """
    df = DataFrame({'int': [1, None], 'float': [.1, None],
                    'complex': [1j, None], 'string': ['a', None]})
    df['category'] = df['string'].astype('category')
    dfm = DataFrameModel(df)
    h, s, v, __ = QColor(dataframeeditor.BACKGROUND_NONNUMBER_COLOR).getHsvF()
    alpha = dataframeeditor.BACKGROUND_MISC_ALPHA
    for idx, column in enumerate(df.columns):
        assert not colorclose(bgcolor(dfm, 0, idx), (h, s, v, alpha)), \
            'Wrong bg color for value of type ' + column
        assert colorclose(bgcolor(dfm, 1, idx), (h, s, v, alpha)), \
            'Wrong bg color for missing of type ' + column


def test_dataframemodel_get_bgcolor_with_nullable_numbers():
    """
    Test background colors for nullable integer data types

    Regression test for spyder-ide/spyder#21222.
    """
    vals = [1, 2, 3, 4, 5]
    vals_na = [1, 2, 3, None, 5]
    df = DataFrame({
        'old': Series(vals),
        'old_na': Series(vals_na),
        'new': Series(vals, dtype='Int64'),
        'new_na': Series(vals_na, dtype='Int64')
    })
    dfm = DataFrameModel(df)
    dfm.colum_avg(0)

    # Test numbers
    h0 = dataframeeditor.BACKGROUND_NUMBER_MINHUE
    dh = dataframeeditor.BACKGROUND_NUMBER_HUERANGE
    s = dataframeeditor.BACKGROUND_NUMBER_SATURATION
    v = dataframeeditor.BACKGROUND_NUMBER_VALUE
    a = dataframeeditor.BACKGROUND_NUMBER_ALPHA
    for col_index in range(4):
        assert colorclose(bgcolor(dfm, 0, col_index), (h0 + dh, s, v, a))
    assert colorclose(bgcolor(dfm, 3, 0), (h0 + 1 / 4 * dh, s, v, a))
    assert colorclose(bgcolor(dfm, 3, 2), (h0 + 1 / 4 * dh, s, v, a))

    # Test null values
    h, s, v, __ = QColor(dataframeeditor.BACKGROUND_NONNUMBER_COLOR).getHsvF()
    alpha = dataframeeditor.BACKGROUND_MISC_ALPHA
    assert colorclose(bgcolor(dfm, 3, 1), (h, s, v, alpha))
    assert colorclose(bgcolor(dfm, 3, 3), (h, s, v, alpha))


def test_dataframemodel_with_format_percent_d_and_nan():
    """
    Test DataFrameModel with format `d` and dataframe containing NaN

    Regression test for spyder-ide/spyder#4139.
    """
    np_array = numpy.zeros(2)
    np_array[1] = numpy.nan
    dataframe = DataFrame(np_array)
    dfm = DataFrameModel(dataframe, format_spec='d')
    assert data(dfm, 0, 0) == '0'
    assert data(dfm, 1, 0) == 'nan'


def test_dataframeeditor_refreshaction_disabled():
    """
    Test that the Refresh action is disabled by default.
    """
    df = DataFrame([[0]])
    editor = DataFrameEditor(None)
    editor.setup_and_check(df)
    assert not editor.refresh_action.isEnabled()


def test_dataframeeditor_refresh():
    """
    Test that after pressing the refresh button, the value of the editor is
    replaced by the return value of the data_function.
    """
    df_zero = DataFrame([[0]])
    df_new = DataFrame([[0, 10], [1, 20], [2, 40]])
    editor = DataFrameEditor(data_function=lambda: df_new)
    editor.setup_and_check(df_zero)
    assert_frame_equal(editor.get_value(), df_zero)
    assert editor.refresh_action.isEnabled()
    editor.refresh_action.trigger()
    assert_frame_equal(editor.get_value(), df_new)


@pytest.mark.parametrize('result', [QMessageBox.Yes, QMessageBox.No])
def test_dataframeeditor_refresh_after_edit(result):
    """
    Test that after changing a value in the editor, pressing the Refresh
    button opens a dialog box (which asks for confirmation), and that the
    editor is only refreshed if the user clicks Yes.
    """
    df_zero = DataFrame([[0]])
    df_edited = DataFrame([[2]])
    df_new = DataFrame([[0, 10], [1, 20], [2, 40]])
    editor = DataFrameEditor(data_function=lambda: df_new)
    editor.setup_and_check(df_zero)
    model = editor.dataModel
    model.setData(model.index(0, 0), '2')
    with patch('spyder.plugins.variableexplorer.widgets.dataframeeditor'
               '.QMessageBox.question',
               return_value=result) as mock_question:
        editor.refresh_action.trigger()
    mock_question.assert_called_once()
    editor.accept()
    if result == QMessageBox.Yes:
        assert_frame_equal(editor.get_value(), df_new)
    else:
        assert_frame_equal(editor.get_value(), df_edited)


def test_dataframeeditor_refresh_into_int(qtbot):
    """
    Test that if the value after refreshing is not a DataFrame but an integer,
    a critical dialog box is displayed and that the editor is closed.
    """
    df_zero = DataFrame([[0]])
    editor = DataFrameEditor(data_function=lambda: 1)
    editor.setup_and_check(df_zero)
    with patch('spyder.plugins.variableexplorer.widgets.dataframeeditor'
               '.QMessageBox.critical') as mock_critical, \
         qtbot.waitSignal(editor.rejected, timeout=0):
        editor.refresh_action.trigger()
    mock_critical.assert_called_once()


def test_dataframeeditor_refresh_when_variable_deleted(qtbot):
    """
    Test that if the variable is deleted and then the editor is refreshed
    (resulting in data_function raising a KeyError), a critical dialog box
    is displayed and that the dataframe editor is closed.
    """
    def datafunc():
        raise KeyError
    df_zero = DataFrame([[0]])
    editor = DataFrameEditor(data_function=datafunc)
    editor.setup_and_check(df_zero)
    with patch('spyder.plugins.variableexplorer.widgets.dataframeeditor'
               '.QMessageBox.critical') as mock_critical, \
         qtbot.waitSignal(editor.rejected, timeout=0):
        editor.refresh_action.trigger()
    mock_critical.assert_called_once()


def test_change_format(qtbot):
    df = DataFrame([[0]])
    editor = DataFrameEditor(None)
    editor.setup_and_check(df)

    def fake_exec(self):
        self.float_format = '10.3e'
        return QDialog.Accepted

    with patch(
        'spyder.plugins.variableexplorer.widgets.dataframeeditor'
        '.PreferencesDialog.exec_',
        fake_exec
    ):
        editor.show_preferences_dialog()

    assert editor.dataModel._format_spec == '10.3e'
    assert editor.get_conf('dataframe_format') == '10.3e'
    editor.set_conf('dataframe_format', '.6g')


def test_dataframemodel_with_format_thousands():
    """
    Check that format can include thousands separator.

    Regression test for spyder-ide/spyder#14518.
    """
    dataframe = DataFrame([10000.1])
    dfm = DataFrameModel(dataframe, format_spec=',.2f')
    assert data(dfm, 0, 0) == '10,000.10'


@flaky(max_runs=3)
def test_dataframeeditor_menu_options(qtbot, monkeypatch):

    def create_view(qtbot, value):
        """Auxiliary function for this test."""
        df = DataFrame(data=value)
        editor = DataFrameEditor()
        assert editor.setup_and_check(df, 'Test DataFrame To action')
        with qtbot.waitExposed(editor):
            editor.show()
        view = editor.dataTable
        dfm = editor.dataModel
        return view, editor, dfm

    d = {'COLUMN_1': [1, 2]}
    view, editor, dfm = create_view(qtbot, d)
    attr_to_patch = ('spyder.plugins.variableexplorer.widgets' +
                     '.dataframeeditor.QMessageBox.question')
    monkeypatch.setattr(attr_to_patch, lambda *args: QMessageBox.Yes)

    # test remove item1 (row)
    view.setCurrentIndex(view.model().index(1, 0))
    assert dfm.rowCount() == 2
    assert dfm.columnCount() == 1
    view.remove_row_action.triggered.emit(True)
    assert editor.btn_save_and_close.isEnabled()
    assert dfm.rowCount() == 1

    # test remove item2 (row)
    view.setCurrentIndex(view.model().index(0, 0))
    view.remove_row_action.triggered.emit(True)
    assert editor.btn_save_and_close.isEnabled()
    assert dfm.rowCount() == 0
    qtbot.mouseClick(editor.btn_save_and_close, Qt.LeftButton)

    # test remove item1 (column)
    d = {'COLUMN_1': [1, 2]}
    view, editor, dfm = create_view(qtbot, d)
    view.setCurrentIndex(view.model().index(0, 0))
    assert dfm.rowCount() == 2
    assert dfm.columnCount() == 1
    view.remove_col_action.triggered.emit(True)
    assert dfm.columnCount() == 0
    assert editor.btn_save_and_close.isEnabled()
    qtbot.mouseClick(editor.btn_save_and_close, Qt.LeftButton)

    # test insert above
    d = {'COLUMN_1': [1, 2, 3], 'COLUMN_2': [4, 5, 6]}
    view, editor, dfm = create_view(qtbot, d)
    view.setCurrentIndex(view.model().index(0, 0))
    assert dfm.rowCount() == 3
    assert dfm.columnCount() == 2
    view.insert_action_above.triggered.emit(True)
    assert dfm.rowCount() == 4
    assert dfm.columnCount() == 2

    # test insert bellow
    view.setCurrentIndex(view.model().index(2, 0))
    view.insert_action_below.triggered.emit(True)
    assert dfm.rowCount() == 5
    assert dfm.columnCount() == 2

    # test insert after
    view.setCurrentIndex(view.model().index(4, 1))
    view.insert_action_after.triggered.emit(True)
    assert dfm.rowCount() == 5
    assert dfm.columnCount() == 3

    # test insert before
    view.setCurrentIndex(view.model().index(4, 0))
    view.insert_action_before.triggered.emit(True)
    assert dfm.rowCount() == 5
    assert dfm.columnCount() == 4

    # duplicate row
    view.setCurrentIndex(view.model().index(0, 3))
    view.duplicate_row_action.triggered.emit(True)
    assert dfm.rowCount() == 6
    assert dfm.columnCount() == 4

    # duplicate column (2x)
    view.setCurrentIndex(view.model().index(1, 3))
    view.duplicate_col_action.triggered.emit(True)
    assert dfm.rowCount() == 6
    assert dfm.columnCount() == 5
    view.setCurrentIndex(view.model().index(0, 1))
    view.duplicate_col_action.triggered.emit(True)
    assert dfm.rowCount() == 6
    assert dfm.columnCount() == 6

    # test edit item
    view.setCurrentIndex(view.model().index(0, 2))
    view.edit_action.triggered.emit(True)
    qtbot.wait(200)
    view.setCurrentIndex(view.model().index(0, 2))
    assert data(dfm, 0, 2) == '0'
    qtbot.keyPress(view.focusWidget(), Qt.Key_9)
    qtbot.keyPress(view.focusWidget(), Qt.Key_Return)
    qtbot.wait(200)
    assert data(dfm, 0, 2) == '9'

    # test edit horizontal header
    monkeypatch.setattr(
        QInputDialog,
        "getText",
        lambda *args: ("SPYDERTEST_H", True)
    )
    header = editor.table_header.model()
    model_index = view.header_class.model().index(0, 2)
    view.header_class.setCurrentIndex(model_index)
    qtbot.wait(200)
    view.menu_header_h.render()
    view.menu_header_h.show()
    qtbot.keyPress(view.menu_header_h, Qt.Key_Down)
    qtbot.keyPress(view.menu_header_h, Qt.Key_Return)
    qtbot.wait(200)
    assert header.headerData(
        2,
        Qt.Horizontal,
        Qt.DisplayRole
    ) == "SPYDERTEST_H"

    # test edit vertical header
    index = editor.table_index.model()
    model_index = editor.table_index.model().index(5, 0)
    editor.table_index.setCurrentIndex(model_index)
    editor.menu_header_v.render()
    editor.menu_header_v.show()
    qtbot.wait(200)
    qtbot.keyPress(editor.menu_header_v, Qt.Key_Down)
    qtbot.keyPress(editor.menu_header_v, Qt.Key_Return)
    qtbot.wait(200)
    qtbot.keyPress(editor.focusWidget(), Qt.Key_9)
    qtbot.keyPress(editor.focusWidget(), Qt.Key_Return)
    qtbot.wait(200)
    assert data_index(index, 5, 0) == '9'
    assert editor.btn_save_and_close.isEnabled()
    qtbot.mouseClick(editor.btn_save_and_close, Qt.LeftButton)


def test_dataframeeditor_with_various_indexes():
    for rng_name, rng in generate_pandas_indexes().items():
        editor = DataFrameEditor(None)
        editor.setup_and_check(rng)
        dfm = editor.dataModel
        assert dfm.rowCount() == 20
        assert dfm.columnCount() == 1
        header = editor.table_header.model()
        assert header.headerData(0, Qt.Horizontal,
                                 Qt.DisplayRole) == "0"

        if rng_name == "Index" and parse(pandas_version) < parse('2.0.0'):
            assert data(dfm, 0, 0) == 'A'
            assert data(dfm, 1, 0) == 'B'
            assert data(dfm, 2, 0) == 'C'
            assert data(dfm, 19, 0) == 'T'
        elif rng_name == "RangeIndex":
            assert data(dfm, 0, 0) == '0'
            assert data(dfm, 1, 0) == '1'
            assert data(dfm, 2, 0) == '2'
            assert data(dfm, 19, 0) == '19'
        elif rng_name in ["Float64Index", "Index"]:
            assert data(dfm, 0, 0) == '0'
            assert data(dfm, 1, 0) == '0.1'
            assert data(dfm, 2, 0) == '0.2'
            assert data(dfm, 19, 0) == '1.9'
        elif rng_name == "DatetimeIndex":
            assert data(dfm, 0, 0) == '2017-01-01 00:00:00'
            assert data(dfm, 1, 0) == '2017-01-02 00:00:00'
            assert data(dfm, 2, 0) == '2017-01-03 00:00:00'
            assert data(dfm, 19, 0) == '2017-01-20 00:00:00'
        elif rng_name == "MultiIndex":
            assert data(dfm, 0, 0) == "('A', 'foo')"
            assert data(dfm, 1, 0) == "('A', 'bar')"
            assert data(dfm, 2, 0) == "('B', 'foo')"
            assert data(dfm, 19, 0) == "('J', 'bar')"
        elif rng_name == "CategoricalIndex":
            assert data(dfm, 0, 0) == 'a'
            assert data(dfm, 1, 0) == 'b'
            assert data(dfm, 2, 0) == 'c'
            assert data(dfm, 19, 0) == 'b'


def test_dataframeeditor_with_OutOfBoundsDatetime():
    # Test for spyder-ide/spyder#6177.
    df = DataFrame([{'DATETIME': datetime.strptime("9999-1-1T00:00",
                                                   "%Y-%m-%dT%H:%M")}])
    model = DataFrameModel(df)
    try:
        model.get_value(0, 0)
    except Exception:
        assert False


@pytest.mark.skipif(not os.name == 'nt',
                    reason="It segfaults too much on Linux")
def test_sort_dataframe_with_duplicate_column(qtbot):
    df = DataFrame({'A': [1, 3, 2], 'B': [4, 6, 5]})
    df = concat((df, df.A), axis=1)
    editor = DataFrameEditor(None)
    editor.setup_and_check(df)
    dfm = editor.dataModel
    QTimer.singleShot(1000, lambda: close_message_box(qtbot))
    editor.dataModel.sort(0)
    assert [data(dfm, row, 0) for row in range(len(df))] == ['1', '3', '2']
    assert [data(dfm, row, 1) for row in range(len(df))] == ['4', '6', '5']
    editor.dataModel.sort(1)
    assert [data(dfm, row, 0) for row in range(len(df))] == ['1', '2', '3']
    assert [data(dfm, row, 1) for row in range(len(df))] == ['4', '5', '6']


@pytest.mark.skipif(not os.name == 'nt',
                    reason="It segfaults too much on Linux")
def test_sort_dataframe_with_category_dtypes(qtbot):
    # cf. spyder-ide/spyder#5361.
    df = DataFrame({'A': [1, 2, 3, 4],
                    'B': ['a', 'b', 'c', 'd']})
    df = df.astype(dtype={'B': 'category'})
    df_cols = df.dtypes
    editor = DataFrameEditor(None)
    editor.setup_and_check(df_cols)
    dfm = editor.dataModel
    QTimer.singleShot(1000, lambda: close_message_box(qtbot))
    editor.dataModel.sort(0)
    assert data(dfm, 0, 0) == 'int64'
    assert data(dfm, 1, 0) == 'category'


def test_dataframemodel_set_data_overflow(monkeypatch):
    """
    Test that entry of an overflowing integer is caught and handled properly.

    Unit regression test for spyder-ide/spyder#6114.
    """
    MockQMessageBox = Mock()
    attr_to_patch = ('spyder.plugins.variableexplorer.widgets' +
                     '.dataframeeditor.QMessageBox')
    monkeypatch.setattr(attr_to_patch, MockQMessageBox)

    # Numpy doesn't raise the OverflowError for ints smaller than 64 bits
    if not os.name == 'nt':
        int32_bit_exponent = 66
    else:
        int32_bit_exponent = 34
    test_parameters = [(1, numpy.int32, int32_bit_exponent),
                       (2, numpy.int64, 66)]

    for idx, int_type, bit_exponent in test_parameters:
        test_df = DataFrame(numpy.arange(7, 11), dtype=int_type)
        model = DataFrameModel(test_df.copy())
        index = model.createIndex(2, 0)
        assert not model.setData(index, str(int(2 ** bit_exponent)))
        MockQMessageBox.critical.assert_called_with(ANY, "Error", ANY)
        assert MockQMessageBox.critical.call_count == idx
        try:
            assert numpy.sum(test_df[0].values ==
                             model.df.values) == len(test_df)
        except AttributeError:
            assert numpy.sum(test_df[0].as_matrix() ==
                             model.df.as_matrix()) == len(test_df)


@flaky(max_runs=3)
@pytest.mark.no_xvfb
@pytest.mark.skipif(sys.platform == 'darwin', reason="It fails on macOS")
def test_dataframeeditor_edit_overflow(qtbot, monkeypatch):
    """
    Test that entry of an overflowing integer is caught and handled properly.

    Integration regression test for spyder-ide/spyder#6114.
    """
    MockQMessageBox = Mock()
    attr_to_patch = ('spyder.plugins.variableexplorer.widgets' +
                     '.dataframeeditor.QMessageBox')
    monkeypatch.setattr(attr_to_patch, MockQMessageBox)

    # Numpy doesn't raise the OverflowError for ints smaller than 64 bits
    if not os.name == 'nt':
        int32_bit_exponent = 66
    else:
        int32_bit_exponent = 34
    test_parameters = [(1, numpy.int32, int32_bit_exponent),
                       (2, numpy.int64, 66)]
    expected_df = DataFrame([5, 6, 7, 3, 4])

    for idx, int_type, bit_exponet in test_parameters:
        test_df = DataFrame(numpy.arange(0, 5), dtype=int_type)
        dialog = DataFrameEditor()
        assert dialog.setup_and_check(test_df, 'Test Dataframe')
        with qtbot.waitExposed(dialog):
            dialog.show()
        view = dialog.dataTable

        qtbot.keyClick(view, Qt.Key_Right)
        qtbot.keyClicks(view, '5')
        qtbot.keyClick(view, Qt.Key_Down)
        qtbot.keyClick(view, Qt.Key_Space)
        qtbot.keyClick(view.focusWidget(), Qt.Key_Backspace)
        qtbot.keyClicks(view.focusWidget(), str(int(2 ** bit_exponet)))
        qtbot.keyClick(view.focusWidget(), Qt.Key_Down)
        MockQMessageBox.critical.assert_called_with(ANY, "Error", ANY)
        assert MockQMessageBox.critical.call_count == idx
        qtbot.keyClicks(view, '7')
        qtbot.keyClick(view, Qt.Key_Up)
        qtbot.keyClicks(view, '6')
        qtbot.keyClick(view, Qt.Key_Down)
        qtbot.wait(200)
        dialog.accept()
        qtbot.wait(500)
        try:
            assert numpy.sum(expected_df[0].values ==
                             dialog.get_value().values) == len(expected_df)
        except AttributeError:
            assert numpy.sum(
                    expected_df[0].as_matrix() ==
                    dialog.get_value().as_matrix()) == len(expected_df)


def test_dataframemodel_set_data_complex(monkeypatch):
    """
    Test that editing complex dtypes is handled gracefully in df editor.

    Unit regression test for spyder-ide/spyder#6115.
    """
    MockQMessageBox = Mock()
    attr_to_patch = ('spyder.plugins.variableexplorer.widgets' +
                     '.dataframeeditor.QMessageBox')
    monkeypatch.setattr(attr_to_patch, MockQMessageBox)

    test_params = [(1, numpy.complex128), (2, numpy.complex64), (3, complex)]

    for count, complex_type in test_params:
        test_df = DataFrame(numpy.arange(10, 15), dtype=complex_type)
        model = DataFrameModel(test_df.copy())
        index = model.createIndex(2, 0)
        assert not model.setData(index, '42')
        MockQMessageBox.critical.assert_called_with(ANY, "Error", ANY)
        assert MockQMessageBox.critical.call_count == count
        try:
            assert numpy.sum(test_df[0].values ==
                             model.df.values) == len(test_df)
        except AttributeError:
            assert numpy.sum(test_df[0].as_matrix() ==
                             model.df.as_matrix()) == len(test_df)


@flaky(max_runs=3)
@pytest.mark.no_xvfb
@pytest.mark.skipif(sys.platform == 'darwin', reason="It fails on macOS")
def test_dataframeeditor_edit_complex(qtbot, monkeypatch):
    """
    Test that editing complex dtypes is handled gracefully in df editor.

    Integration regression test for spyder-ide/spyder#6115.
    """
    MockQMessageBox = Mock()
    attr_to_patch = ('spyder.plugins.variableexplorer.widgets' +
                     '.dataframeeditor.QMessageBox')
    monkeypatch.setattr(attr_to_patch, MockQMessageBox)

    test_params = [(1, numpy.complex128), (2, numpy.complex64), (3, complex)]

    for count, complex_type in test_params:
        test_df = DataFrame(numpy.arange(10, 15), dtype=complex_type)
        dialog = DataFrameEditor()
        assert dialog.setup_and_check(test_df, 'Test Dataframe')
        with qtbot.waitExposed(dialog):
            dialog.show()
        view = dialog.dataTable

        qtbot.keyClick(view, Qt.Key_Right)
        qtbot.keyClick(view, Qt.Key_Down)
        qtbot.keyClick(view, Qt.Key_Space)
        qtbot.keyClick(view.focusWidget(), Qt.Key_Backspace)
        qtbot.keyClicks(view.focusWidget(), "42")
        qtbot.keyClick(view.focusWidget(), Qt.Key_Down)
        MockQMessageBox.critical.assert_called_with(ANY, "Error", ANY)
        assert MockQMessageBox.critical.call_count == count * 2 - 1
        qtbot.keyClick(view, Qt.Key_Down)
        qtbot.keyClick(view, '1')
        qtbot.keyClick(view.focusWidget(), Qt.Key_Down)
        MockQMessageBox.critical.assert_called_with(
            ANY, "Error", ("Editing dtype {0!s} not yet supported."
                           .format(type(test_df.iloc[1, 0]).__name__)))
        assert MockQMessageBox.critical.call_count == count * 2
        qtbot.wait(200)
        dialog.accept()
        qtbot.wait(500)
        try:
            assert numpy.sum(test_df[0].values ==
                             dialog.get_value().values) == len(test_df)
        except AttributeError:
            assert numpy.sum(test_df[0].as_matrix() ==
                             dialog.get_value().as_matrix()) == len(test_df)


def test_dataframemodel_set_data_bool(monkeypatch):
    """Test that bools are editible in df and false-y strs are detected."""
    MockQMessageBox = Mock()
    attr_to_patch = ('spyder.plugins.variableexplorer.widgets' +
                     '.dataframeeditor.QMessageBox')
    monkeypatch.setattr(attr_to_patch, MockQMessageBox)

    test_params = [numpy.bool_, bool]
    test_strs = ['foo', 'false', 'f', '0', '0.', '0.0', '', ' ']
    expected_df = DataFrame([1, 0, 0, 0, 0, 0, 0, 0, 0], dtype=bool)

    for bool_type in test_params:
        test_df = DataFrame([0, 1, 1, 1, 1, 1, 1, 1, 0], dtype=bool_type)
        model = DataFrameModel(test_df.copy())
        for idx, test_str in enumerate(test_strs):
            assert model.setData(model.createIndex(idx, 0), test_str)
            assert not MockQMessageBox.critical.called
        try:
            assert numpy.sum(expected_df[0].values ==
                             model.df.values[:, 0]) == len(expected_df)
        except AttributeError:
            assert numpy.sum(expected_df[0].as_matrix() ==
                             model.df.as_matrix()[:, 0]) == len(expected_df)


@flaky(max_runs=3)
@pytest.mark.no_xvfb
@pytest.mark.skipif(sys.platform == 'darwin', reason="It fails on macOS")
def test_dataframeeditor_edit_bool(qtbot, monkeypatch):
    """Test that bools are editible in df and false-y strs are detected."""
    MockQMessageBox = Mock()
    attr_to_patch = ('spyder.plugins.variableexplorer.widgets' +
                     '.dataframeeditor.QMessageBox')
    monkeypatch.setattr(attr_to_patch, MockQMessageBox)

    test_params = [numpy.bool_, bool]
    test_strs = ['foo', 'false', 'f', '0', '0.', '0.0', '', ' ']
    expected_df = DataFrame([1, 0, 0, 0, 0, 0, 0, 0, 0], dtype=bool)

    for bool_type in test_params:
        test_df = DataFrame([0, 1, 1, 1, 1, 1, 1, 1, 0], dtype=bool_type)
        dialog = DataFrameEditor()
        assert dialog.setup_and_check(test_df, 'Test Dataframe')
        with qtbot.waitExposed(dialog):
            dialog.show()
        view = dialog.dataTable

        qtbot.keyClick(view, Qt.Key_Right)
        for test_str in test_strs:
            qtbot.keyClick(view, Qt.Key_Space)
            qtbot.keyClick(view.focusWidget(), Qt.Key_Backspace)
            qtbot.keyClicks(view.focusWidget(), test_str)
            qtbot.keyClick(view.focusWidget(), Qt.Key_Down)
            assert not MockQMessageBox.critical.called
        qtbot.wait(200)
        dialog.accept()
        qtbot.wait(500)
        try:
            assert (numpy.sum(expected_df[0].values ==
                              dialog.get_value().values[:, 0]) ==
                    len(expected_df))
        except AttributeError:
            assert (numpy.sum(expected_df[0].as_matrix() ==
                              dialog.get_value().as_matrix()[:, 0]) ==
                    len(expected_df))


def test_non_ascii_index():
    """
    Test that there are no errors when displaying a dataframe with
    a non-ascii index and header.
    """
    df = read_csv(os.path.join(FILES_PATH, 'issue_5833.csv'), index_col=0)
    editor = DataFrameEditor(None)
    editor.setup_and_check(df)

    index = editor.table_index.model()
    header = editor.table_header.model()
    dfm = editor.model()

    assert header.headerData(0, Qt.Horizontal,
                             Qt.DisplayRole) == "кодирование"
    assert data_index(index, 0, 0) == 'пример'
    assert data(dfm, 0, 0) == 'файла'


def test_no_convert_strings_to_unicode():
    """
    Test that we don't apply any conversion to strings in headers,
    indexes or data.
    """
    df = read_csv(os.path.join(FILES_PATH, 'issue_5833.csv'), index_col=0,
                  encoding='koi8_r')
    editor = DataFrameEditor(None)
    editor.setup_and_check(df)

    index = editor.table_index.model()
    header = editor.table_header.model()
    dfm = editor.model()

    assert header.headerData(0, Qt.Horizontal,
                             Qt.DisplayRole) != u"кодирование"
    assert data_index(index, 0, 0) != u'пример'
    assert data(dfm, 0, 0) != u'файла'


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Dialog window for recovering files from autosave"""

# Standard library imports
from os import path as osp
import os
import shutil
import time

# Third party imports
from qtpy.compat import getsavefilename
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QTableWidget,
    QVBoxLayout,
    QWidget,
)

# Local imports
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.config.base import _, running_under_pytest


class RecoveryDialog(QDialog):
    """Dialog window to allow users to recover from autosave files."""
    def __init__(self, autosave_mapping, parent=None):
        """
        Constructor

        Parameters
        ----------
        autosave_mapping : List[Tuple[str]]
            List of tuples, containing the name of the original file and the
            name of the corresponding autosave file. The first entry of the
            tuple may be `None` to indicate that the original file is unknown.
        parent : QWidget, optional
            Parent of the dialog window. The default is None.
        """
        QDialog.__init__(self, parent)
        self.layout = QVBoxLayout(self)
        self.setLayout(self.layout)
        self.layout.setSpacing(self.layout.spacing() * 3)
        self.files_to_open = []
        self.gather_data(autosave_mapping)
        self.add_label()
        self.add_table()
        self.add_cancel_button()
        self.setWindowTitle(_('Recover from autosave'))
        self.setFixedSize(670, 400)
        self.setWindowFlags(
            Qt.Dialog | Qt.MSWindowsFixedSizeDialogHint |
            Qt.WindowStaysOnTopHint)

        # This is needed because of an error in MacOS.
        # See https://bugreports.qt.io/browse/QTBUG-49576
        if parent and hasattr(parent, 'splash'):
            self.splash = parent.splash
            self.splash.hide()
        else:
            self.splash = None

    def accept(self):
        """Reimplement Qt method."""
        if self.splash is not None:
            self.splash.show()
        super(RecoveryDialog, self).accept()

    def reject(self):
        """Reimplement Qt method."""
        if self.splash is not None:
            self.splash.show()
        super(RecoveryDialog, self).reject()

    def gather_file_data(self, name):
        """
        Gather data about a given file.

        Returns a dict with fields 'name', 'mtime' and 'size', containing the
        relevant data for the file. If the file does not exists, then the dict
        contains only the field `name`.
        """
        res = {'name': name}
        try:
            res['mtime'] = osp.getmtime(name)
            res['size'] = osp.getsize(name)
        except OSError:
            pass
        return res

    def gather_data(self, autosave_mapping):
        """
        Gather data about files which may be recovered.

        The data is stored in self.data as a list of tuples with the data
        pertaining to the original file and the autosave file. Each element of
        the tuple is a dict as returned by gather_file_data(). Autosave files
        which do not exist, are ignored.
        """
        self.data = []
        for orig, autosave in autosave_mapping:
            if orig:
                orig_dict = self.gather_file_data(orig)
            else:
                orig_dict = None
            autosave_dict = self.gather_file_data(autosave)
            if 'mtime' not in autosave_dict:  # autosave file does not exist
                continue
            self.data.append((orig_dict, autosave_dict))
        self.data.sort(key=self.recovery_data_key_function)
        self.num_enabled = len(self.data)

    def recovery_data_key_function(self, item):
        """
        Convert item in `RecoveryDialog.data` to tuple so that it can be sorted.

        Sorting the tuples returned by this function will sort first by name of
        the original file, then by name of the autosave file. All items without an
        original file name will be at the end.
        """
        orig_dict, autosave_dict = item
        if orig_dict:
            return (0, orig_dict['name'], autosave_dict['name'])
        else:
            return (1, 0, autosave_dict['name'])

    def add_label(self):
        """Add label with explanation at top of dialog window."""
        txt = _('Autosave files found. What would you like to do?\n\n'
                'This dialog will be shown again on next startup if any '
                'autosave files are not restored, moved or deleted.')
        label = QLabel(txt, self)
        label.setWordWrap(True)
        self.layout.addWidget(label)

    def add_label_to_table(self, row, col, txt):
        """Add a label to specified cell in table."""
        label = QLabel(txt)
        label.setMargin(5)
        label.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        self.table.setCellWidget(row, col, label)

    def add_table(self):
        """Add table with info about files to be recovered."""
        table = QTableWidget(len(self.data), 3, self)
        self.table = table

        labels = [_('Original file'), _('Autosave file'), _('Actions')]
        table.setHorizontalHeaderLabels(labels)
        table.verticalHeader().hide()

        table.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        table.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        table.setSelectionMode(QTableWidget.NoSelection)

        # Show horizontal grid lines
        table.setShowGrid(False)
        table.setStyleSheet('::item { border-bottom: 1px solid gray }')

        for idx, (original, autosave) in enumerate(self.data):
            self.add_label_to_table(idx, 0, self.file_data_to_str(original))
            self.add_label_to_table(idx, 1, self.file_data_to_str(autosave))

            widget = QWidget()
            layout = QHBoxLayout()

            tooltip = _('Recover the autosave file to its original location, '
                        'replacing the original if it exists.')
            button = QPushButton(_('Restore'))
            button.setToolTip(tooltip)
            button.clicked[bool].connect(
                    lambda checked, my_idx=idx: self.restore(my_idx))
            layout.addWidget(button)

            tooltip = _('Delete the autosave file.')
            button = QPushButton(_('Discard'))
            button.setToolTip(tooltip)
            button.clicked[bool].connect(
                    lambda checked, my_idx=idx: self.discard(my_idx))
            layout.addWidget(button)

            tooltip = _('Display the autosave file (and the original, if it '
                        'exists) in Spyder\'s Editor. You will have to move '
                        'or delete it manually.')
            button = QPushButton(_('Open'))
            button.setToolTip(tooltip)
            button.clicked[bool].connect(
                    lambda checked, my_idx=idx: self.open_files(my_idx))
            layout.addWidget(button)

            widget.setLayout(layout)
            self.table.setCellWidget(idx, 2, widget)

        table.resizeRowsToContents()
        table.resizeColumnsToContents()

        self.layout.addWidget(table)

    def file_data_to_str(self, data):
        """
        Convert file data to a string for display.

        This function takes the file data produced by gather_file_data().
        """
        if not data:
            return _('<i>File name not recorded</i>')
        res = data['name']
        try:
            mtime_as_str = time.strftime('%Y-%m-%d %H:%M:%S',
                                        time.localtime(data['mtime']))
            res += '<br><i>{}</i>: {}'.format(_('Last modified'), mtime_as_str)
            res += u'<br><i>{}</i>: {} {}'.format(
                    _('Size'), data['size'], _('bytes'))
        except KeyError:
            res += '<br>' + _('<i>File no longer exists</i>')
        return res

    def add_cancel_button(self):
        """Add a cancel button at the bottom of the dialog window."""
        button_box = SpyderDialogButtonBox(
            QDialogButtonBox.Cancel, parent=self
        )
        button_box.rejected.connect(self.reject)
        self.layout.addWidget(button_box)

    def center(self):
        """Center the dialog."""
        screen = self.screen().geometry()
        x = int(screen.center().x() - self.width() / 2)
        y = int(screen.center().y() - self.height() / 2)
        self.move(x, y)

    def restore(self, idx):
        orig, autosave = self.data[idx]
        if orig:
            orig_name = orig['name']
        else:
            orig_name, ignored = getsavefilename(
                self, _('Restore autosave file to ...'),
                osp.basename(autosave['name']))
            if not orig_name:
                return
        try:
            try:
                os.replace(autosave['name'], orig_name)
            except (AttributeError, OSError):
                # os.replace() does not exist on Python 2 and fails if the
                # files are on different file systems.
                # See spyder-ide/spyder#8631.
                shutil.copy2(autosave['name'], orig_name)
                os.remove(autosave['name'])
            self.deactivate(idx)
        except EnvironmentError as error:
            text = (_('Unable to restore {} using {}')
                    .format(orig_name, autosave['name']))
            self.report_error(text, error)

    def discard(self, idx):
        ignored, autosave = self.data[idx]
        try:
            os.remove(autosave['name'])
            self.deactivate(idx)
        except EnvironmentError as error:
            text = _('Unable to discard {}').format(autosave['name'])
            self.report_error(text, error)

    def open_files(self, idx):
        orig, autosave = self.data[idx]
        if orig:
            self.files_to_open.append(orig['name'])
        self.files_to_open.append(autosave['name'])
        self.deactivate(idx)

    def report_error(self, text, error):
        heading = _('Error message:')
        msgbox = QMessageBox(
            QMessageBox.Critical, _('Restore'),
            _('<b>{}</b><br><br>{}<br>{}').format(text, heading, error),
            parent=self)
        msgbox.exec_()

    def deactivate(self, idx):
        for col in range(self.table.columnCount()):
            self.table.cellWidget(idx, col).setEnabled(False)
        self.num_enabled -= 1
        if self.num_enabled == 0:
            self.accept()

    def exec_if_nonempty(self):
        """Execute dialog window if there is data to show."""
        if self.data:
            self.center()
            return self.exec_()
        else:
            return QDialog.Accepted

    def exec_(self):
        """Execute dialog window."""
        if running_under_pytest():
            return QDialog.Accepted
        return super(RecoveryDialog, self).exec_()


def make_temporary_files(tempdir):
    """
    Make temporary files to simulate a recovery use case.

    Create a directory under tempdir containing some original files and another
    directory with autosave files. Return a tuple with the name of the
    directory with the original files, the name of the directory with the
    autosave files, and the autosave mapping.
    """
    orig_dir = osp.join(tempdir, 'orig')
    os.mkdir(orig_dir)
    autosave_dir = osp.join(tempdir, 'autosave')
    os.mkdir(autosave_dir)
    autosave_mapping = {}

    # ham.py: Both original and autosave files exist, mentioned in mapping
    orig_file = osp.join(orig_dir, 'ham.py')
    with open(orig_file, 'w') as f:
        f.write('ham = "original"\n')
    autosave_file = osp.join(autosave_dir, 'ham.py')
    with open(autosave_file, 'w') as f:
        f.write('ham = "autosave"\n')
    autosave_mapping = [(orig_file, autosave_file)]

    # spam.py: Only autosave file exists, mentioned in mapping
    orig_file = osp.join(orig_dir, 'spam.py')
    autosave_file = osp.join(autosave_dir, 'spam.py')
    with open(autosave_file, 'w') as f:
        f.write('spam = "autosave"\n')
    autosave_mapping += [(orig_file, autosave_file)]

    # eggs.py: Only original files exists, mentioned in mapping
    orig_file = osp.join(orig_dir, 'eggs.py')
    with open(orig_file, 'w') as f:
        f.write('eggs = "original"\n')
    autosave_file = osp.join(autosave_dir, 'eggs.py')
    autosave_mapping += [(orig_file, autosave_file)]

    # cheese.py: Only autosave file exists
    autosave_file = osp.join(autosave_dir, 'cheese.py')
    with open(autosave_file, 'w') as f:
        f.write('cheese = "autosave"\n')
    autosave_mapping += [(None, autosave_file)]

    return orig_dir, autosave_dir, autosave_mapping


def test():  # pragma: no cover
    """Display recovery dialog for manual testing."""
    import shutil
    import tempfile
    from spyder.utils.qthelpers import qapplication

    app = qapplication()  # noqa
    tempdir = tempfile.mkdtemp()
    unused, unused, autosave_mapping = make_temporary_files(tempdir)
    dialog = RecoveryDialog(autosave_mapping)
    dialog.exec_()
    print('files_to_open =', dialog.files_to_open)  # spyder: test-skip
    shutil.rmtree(tempdir)


if __name__ == "__main__":  # pragma: no cover
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for browser.py
"""

# Test library imports
import pytest

# Local imports
from spyder.widgets.browser import WebBrowser


@pytest.fixture
def browser(qtbot):
    """Set up WebBrowser."""
    WebBrowser.CONF_SECTION = ''
    widget = WebBrowser(None)
    widget.show()
    yield widget
    widget.webview.stop()
    widget.close()


def test_browser(browser):
    """Run web browser."""
    browser.set_home_url('https://www.google.com/')
    browser.go_home()
    assert browser


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tours Plugin.
"""

# Local imports
from spyder.api.plugins import Plugins, SpyderPluginV2
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.translations import _
from spyder.config.base import get_safe_mode, running_under_pytest
from spyder.plugins.application.api import ApplicationActions
from spyder.plugins.mainmenu.api import ApplicationMenus, HelpMenuSections
from spyder.plugins.tours.container import TourActions, ToursContainer
from spyder.plugins.tours.tours import INTRO_TOUR, TourIdentifiers
from spyder.plugins.tours.api import SpyderWidgets


class Tours(SpyderPluginV2):
    """
    Tours Plugin.
    """
    NAME = 'tours'
    CONF_SECTION = NAME
    OPTIONAL = [Plugins.MainMenu]
    CONF_FILE = False
    CONTAINER_CLASS = ToursContainer

    # ---- SpyderPluginV2 API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _("Interactive tours")

    @staticmethod
    def get_description():
        return _("Provide interactive tours of the Spyder interface.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon('tour')

    def on_initialize(self):
        pass

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_main_menu_available(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)

        mainmenu.add_item_to_application_menu(
            self.get_container().tour_action,
            menu_id=ApplicationMenus.Help,
            section=HelpMenuSections.Documentation,
            before_section=HelpMenuSections.Support,
            before=ApplicationActions.SpyderDocumentationAction)

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_main_menu_teardown(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)
        mainmenu.remove_item_from_application_menu(
            TourActions.ShowTour,
            menu_id=ApplicationMenus.Help)

    def on_mainwindow_visible(self):
        # Remove from intro tour steps for unavailable plugins.
        # Fixes spyder-ide/spyder#22635
        trimmed_intro_tour = self._trim_intro_tour()

        # Register trimmed tour
        self.register_tour(
            TourIdentifiers.IntroductionTour,
            _("Introduction to Spyder"),
            trimmed_intro_tour,
        )

        # Show tour message (only the first time Spyder starts)
        self.show_tour_message()

    # ---- Public API
    # -------------------------------------------------------------------------
    def register_tour(self, tour_id, title, tour_data):
        """
        Register a new interactive tour on spyder.

        Parameters
        ----------
        tour_id: str
            Unique tour string identifier.
        title: str
            Localized tour name.
        tour_data: dict
            The tour steps.
        """
        self.get_container().register_tour(tour_id, title, tour_data)

    def show_tour(self, index):
        """
        Show interactive tour.

        Parameters
        ----------
        index: int
            The tour index to display.
        """
        self.get_container().show_tour(index)

    def show_tour_message(self, force=False):
        """
        Show message about starting the tour the first time Spyder starts.

        Parameters
        ----------
        force: bool
            Force the display of the tour message.
        """
        should_show_tour = self.get_conf('show_tour_message')
        if force or (should_show_tour and not running_under_pytest()
                     and not get_safe_mode()):
            self.set_conf('show_tour_message', False)
            self.get_container().show_tour_message()

    # ---- Private API
    # -------------------------------------------------------------------------
    def _trim_intro_tour(self):
        """Trim intro tour according to available plugins."""
        # Mapping of SpyderWidgets to Plugins (some names are different)
        # TODO: This should be fixed.
        widgets_to_plugins = {
            SpyderWidgets.editor: Plugins.Editor,
            SpyderWidgets.ipython_console: Plugins.IPythonConsole,
            SpyderWidgets.variable_explorer: Plugins.VariableExplorer,
            SpyderWidgets.help_plugin: Plugins.Help,
            SpyderWidgets.plots_plugin: Plugins.Plots,
            SpyderWidgets.file_explorer: Plugins.Explorer,
            SpyderWidgets.history_log: Plugins.History,
            SpyderWidgets.find_plugin: Plugins.Find,
            SpyderWidgets.profiler: Plugins.Profiler,
            SpyderWidgets.code_analysis: Plugins.Pylint,
        }

        # Only leave on tour the steps for available plugins
        trimmed_tour = []
        for step in INTRO_TOUR:
            if "widgets" in step:
                widget = step["widgets"][0]
                if self.is_plugin_available(widgets_to_plugins[widget]):
                    trimmed_tour.append(step)
            else:
                trimmed_tour.append(step)

        return trimmed_tour



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tests for the Projects plugin.
"""

# Standard library imports
import configparser
import os
import os.path as osp
import shutil
import sys
from unittest.mock import MagicMock

# Third party imports
import pytest
from flaky import flaky

# Local imports
from spyder.app.cli_options import get_options
from spyder.config.manager import CONF
import spyder.plugins.base
from spyder.plugins.preferences.tests.conftest import MainWindowMock
from spyder.plugins.projects.api import BaseProjectType
from spyder.plugins.projects.plugin import Projects
from spyder.plugins.projects.widgets.main_widget import QMessageBox
from spyder.plugins.projects.widgets.projectdialog import ProjectDialog
from spyder.py3compat import to_text_string


# =============================================================================
# ---- Fixtures
# =============================================================================
@pytest.fixture
def projects(qtbot, mocker, request, tmpdir):
    """Projects plugin fixture."""
    use_cli_project = request.node.get_closest_marker('use_cli_project')

    class EditorMock(MagicMock):
        def get_open_filenames(self):
            # Patch this with mocker to return a different value.
            # See test_set_project_filenames_in_close_project.
            return []

    class MainWindowProjectsMock(MainWindowMock):
        def __init__(self, parent):
            # This avoids using the cli options passed to pytest
            sys_argv = [sys.argv[0]]
            self._cli_options = get_options(sys_argv)[0]
            super().__init__(parent)

        def __getattr__(self, attr):
            if attr == 'ipyconsole':
                return None
            try:
                super().__getattr__(attr)
            except AttributeError:
                return MagicMock()

        def get_initial_working_directory(self):
            return str(tmpdir)

        def get_plugin(self, plugin_name, error=True):
            if plugin_name == 'editor':
                return EditorMock()
            else:
                return None

    # Main window mock
    main_window = MainWindowProjectsMock(None)
    if use_cli_project:
        tmpdir.mkdir('cli_project_dir')

        # This allows us to test relative paths passed on the command line
        main_window._cli_options.project = 'cli_project_dir'

    # Create plugin
    projects = Projects(parent=None, configuration=CONF)
    projects.initialize()

    projects.sig_switch_to_plugin_requested.connect(
        lambda x, y: projects.change_visibility(True))

    # This can only be done at this point
    projects._main = main_window

    # Patching necessary to test visible_if_project_open
    projects.shortcut = None
    mocker.patch.object(spyder.plugins.base.SpyderDockWidget,
                        'install_tab_event_filter')
    yield projects
    projects.get_container().close()
    projects.on_close()


@pytest.fixture
def create_projects(projects, mocker):
    """Create a Projects plugin fixture"""
    def _create_projects(path, files):
        """
        Using the Projects plugin fixture, open a project at the
        specified path, and mock the opening of the specified files
        in the Editor.
        """
        # Open a project.
        projects.open_project(path=path)

        # Mock the opening of files in the Editor while the project is open.
        projects._get_open_filenames = lambda *args, **kwargs: files
        return projects

    return _create_projects


# =============================================================================
# ---- Tests
# =============================================================================
@pytest.mark.parametrize("test_directory", [u'測試', u'اختبار', u"test_dir"])
def test_open_project(projects, tmpdir, test_directory):
    """Test that we can create a project in a given directory."""
    # Create the directory
    path = to_text_string(tmpdir.mkdir(test_directory))

    # Open project in path
    projects.open_project(path=path)

    # Verify that we created a valid project
    assert projects.is_valid_project(path)

    # Close project
    projects.close_project()


@pytest.mark.parametrize("test_directory", [u'測試', u'اختبار', u"test_dir"])
def test_delete_project(projects, tmpdir, mocker, test_directory):
    """Test that we can delete a project."""
    # Create the directory
    path = to_text_string(tmpdir.mkdir(test_directory))

    # Open project in path
    projects.open_project(path=path)
    assert projects.is_valid_project(path)
    assert osp.exists(osp.join(path, '.spyproject'))

    # Delete project
    mocker.patch.object(QMessageBox, 'warning', return_value=QMessageBox.Yes)
    projects.delete_project()
    assert not projects.is_valid_project(path)
    assert not osp.exists(osp.join(path, '.spyproject'))


@pytest.mark.parametrize('value', [True, False])
def test_close_project_sets_visible_config(projects, tmpdir, value):
    """Test that when project is closed, the config option
    visible_if_project_open is set to the correct value."""
    # Set config to opposite value so that we can check that it's set correctly
    projects.set_conf('visible_if_project_open', not value)
    projects.open_project(path=to_text_string(tmpdir))
    if value:
        projects._show_main_widget()
    else:
        projects.get_widget().hide()
    projects.close_project()
    assert projects.get_conf('visible_if_project_open') == value


@pytest.mark.parametrize('value', [True, False])
def test_on_close_sets_visible_config(projects, tmpdir, value):
    """Test that on_close() sets config option visible_if_project_open
    if a project is open."""
    projects.set_conf('visible_if_project_open', not value)
    projects.on_close()

    # No project is open so config option should remain unchanged
    assert projects.get_conf('visible_if_project_open') == (not value)

    projects.open_project(path=to_text_string(tmpdir))
    if value:
        projects._show_main_widget()
    else:
        projects.get_widget().hide()
    projects.close_project()
    assert projects.get_conf('visible_if_project_open') == value


@pytest.mark.parametrize('value', [True, False])
def test_open_project_uses_visible_config(projects, tmpdir, value):
    """Test that when a project is opened, the project explorer is only opened
    if the config option visible_if_project_open is set."""
    projects.set_conf('visible_if_project_open', value)
    projects.open_project(path=to_text_string(tmpdir))
    assert projects.get_widget().isVisible() == value


@pytest.mark.parametrize('value', [False, True])
def test_switch_to_plugin(projects, tmpdir, value):
    """Test that switch_to_plugin always shows the plugin if a project is
    opened, regardless of the config option visible_if_project_open.
    Regression test for spyder-ide/spyder#12491."""
    projects.set_conf('visible_if_project_open', value)
    projects.open_project(path=to_text_string(tmpdir))
    projects.switch_to_plugin()
    assert projects.get_widget().isVisible()


def test_set_get_project_filenames_when_closing_no_files(create_projects,
                                                         tmpdir):
    """
    Test that the currently opened files in the Editor are saved and loaded
    correctly to and from the project config when the project is closed and
    then reopened.

    Regression test for spyder-ide/spyder#10045.
    """
    path = to_text_string(tmpdir.mkdir('project1'))
    # Create paths but no actual files
    opened_files = [os.path.join(path, file)
                    for file in ['file1', 'file2', 'file3']]

    # Create the projects plugin without creating files.
    projects = create_projects(path, opened_files)
    assert projects.get_project_filenames() == []

    # Close and reopen the project.
    projects.close_project()
    projects.open_project(path=path)

    # Check the original list isn't empty (was not changed)
    assert opened_files


def test_set_get_project_filenames_when_closing(create_projects, tmpdir):
    """
    Test that the currently opened files in the Editor are saved and loaded
    correctly to and from the project config when the project is closed and
    then reopened.

    Regression test for spyder-ide/spyder#8375.
    Updated for spyder-ide/spyder#10045.
    """
    # Setup tmp dir and files
    dir_object = tmpdir.mkdir('project1')
    path = to_text_string(dir_object)

    # Needed to actually create the files
    opened_files = []
    for file in ['file1', 'file2', 'file3']:
        file_object = dir_object.join(file)
        file_object.write(file)
        opened_files.append(to_text_string(file_object))

    # Create the projects plugin.
    projects = create_projects(path, opened_files)
    assert projects.get_project_filenames() == []

    # Close and reopen the project.
    projects.close_project()
    projects.open_project(path=path)
    assert projects.get_project_filenames() == opened_files


def test_set_get_project_filenames_when_switching(create_projects, tmpdir):
    """
    Test that files in the Editor are loaded and saved correctly when
    switching projects.

    Updated for spyder-ide/spyder#10045.
    """
    dir_object1 = tmpdir.mkdir('project1')
    path1 = to_text_string(dir_object1)
    path2 = to_text_string(tmpdir.mkdir('project2'))

    # Needed to actually create the files
    opened_files = []
    for file in ['file1', 'file2', 'file3']:
        file_object = dir_object1.join(file)
        file_object.write(file)
        opened_files.append(to_text_string(file_object))

    # Create the projects plugin.
    projects = create_projects(path1, opened_files)
    assert projects.get_project_filenames() == []

    # Switch to another project.
    projects.open_project(path=path2)
    assert projects.get_project_filenames() == []

    # Switch back to the first project.
    projects.close_project()
    projects.open_project(path=path1)
    assert projects.get_project_filenames() == opened_files


def test_recent_projects_menu_action(projects, tmpdir):
    """
    Test that the actions of the submenu 'Recent Projects' in the 'Projects'
    main menu are working as expected.

    Regression test for spyder-ide/spyder#8450.
    """
    # Create the directories.
    path0 = to_text_string(tmpdir.mkdir('project0'))
    path1 = to_text_string(tmpdir.mkdir('project1'))
    path2 = to_text_string(tmpdir.mkdir('project2'))

    # Open projects in path0, path1, and path2.
    projects.open_project(path=path0)
    projects.open_project(path=path1)
    projects.open_project(path=path2)
    actions = list(projects.get_widget().get_actions().keys())
    assert path0 in actions
    assert path1 in actions
    assert path2 in actions
    assert projects.get_active_project().root_path == path2

    # Trigger project1 in the list of Recent Projects actions.
    projects.get_widget().get_actions()[path1].trigger()
    assert projects.get_active_project().root_path == path1

    # Trigger project0 in the list of Recent Projects actions.
    projects.get_widget().get_actions()[path0].trigger()
    assert projects.get_active_project().root_path == path0


def test_project_explorer_tree_root(projects, tmpdir, qtbot):
    """
    Test that the root item of the project explorer tree widget is set
    correctly when switching projects.

    Regression test for spyder-ide/spyder#8455.
    """
    qtbot.addWidget(projects.get_widget())
    projects._show_main_widget()

    ppath1 = to_text_string(tmpdir.mkdir(u'測試'))
    ppath2 = to_text_string(tmpdir.mkdir(u'ïèô éàñ').mkdir(u'اختبار'))
    if os.name == 'nt':
        # For an explanation of why this part is necessary to make this test
        # pass for Python2 in Windows, see spyder-ide/spyder#8528.
        import win32file
        ppath1 = win32file.GetLongPathName(ppath1)
        ppath2 = win32file.GetLongPathName(ppath2)

    # Open the projects.
    for ppath in [ppath1, ppath2]:
        projects.open_project(path=ppath)
        projects.get_widget()._setup_project(ppath)

        # Check that the root path of the project explorer tree widget is
        # set correctly.
        assert projects.get_active_project_path() == ppath
        assert projects.get_widget().treewidget.root_path == osp.dirname(ppath)
        assert (projects.get_widget().treewidget.rootIndex().data() ==
                osp.basename(osp.dirname(ppath)))

        # Check that the first visible item in the project explorer
        # tree widget is the folder of the project.
        topleft_index = (projects.get_widget().treewidget.indexAt(
            projects.get_widget().treewidget.rect().topLeft()))
        assert topleft_index.data() == osp.basename(ppath)


@flaky(max_runs=5)
def test_filesystem_notifications(qtbot, projects, tmpdir):
    """
    Test that filesystem notifications are emitted when creating,
    deleting and moving files and directories.
    """
    # Create a directory for the project and some files.
    project_root = tmpdir.mkdir('project0')
    git_folder = project_root.mkdir('.git')
    folder0 = project_root.mkdir('folder0')
    folder1 = project_root.mkdir('folder1')
    file0 = project_root.join('file0.txt')
    file1 = folder0.join('file1.txt')
    file2 = folder0.join('file2.txt')
    file3 = folder1.join('file3.txt')
    file0.write('')
    file1.write('')
    file3.write('ab')

    # Open the project
    projects.open_project(path=str(project_root))

    # Get a reference to the filesystem event handler
    fs_handler = projects.get_widget().watcher.event_handler

    # Test file creation
    with qtbot.waitSignal(fs_handler.sig_file_created,
                          timeout=30000) as blocker:
        file2.write('')

    file_created, is_dir = blocker.args
    assert file_created == str(file2)
    assert not is_dir

    # Test folder creation
    with qtbot.waitSignal(fs_handler.sig_file_created,
                          timeout=3000) as blocker:
        folder2 = project_root.mkdir('folder2')

    folder_created, is_dir = blocker.args
    assert folder_created == osp.join(str(project_root), 'folder2')

    # Test file move/renaming
    new_file = osp.join(str(folder0), 'new_file.txt')
    with qtbot.waitSignal(fs_handler.sig_file_moved,
                          timeout=3000) as blocker:
        shutil.move(str(file1), new_file)

    original_file, file_moved, is_dir = blocker.args
    assert original_file == str(file1)
    assert file_moved == new_file
    assert not is_dir

    # Test folder move/renaming
    new_folder = osp.join(str(project_root), 'new_folder')
    with qtbot.waitSignal(fs_handler.sig_file_moved,
                          timeout=3000) as blocker:
        shutil.move(str(folder2), new_folder)

    original_folder, folder_moved, is_dir = blocker.args
    assert original_folder == str(folder2)
    assert folder_moved == new_folder
    assert is_dir

    # Test file deletion
    with qtbot.waitSignal(fs_handler.sig_file_deleted,
                          timeout=3000) as blocker:
        os.remove(str(file0))

    deleted_file, is_dir = blocker.args
    assert deleted_file == str(file0)
    assert not is_dir
    assert not osp.exists(str(file0))

    # Test folder deletion
    with qtbot.waitSignal(fs_handler.sig_file_deleted,
                          timeout=3000) as blocker:
        shutil.rmtree(str(folder0))

    deleted_folder, is_dir = blocker.args
    assert str(folder0) in deleted_folder

    # Test file/folder modification
    with qtbot.waitSignal(fs_handler.sig_file_modified,
                          timeout=3000) as blocker:
        file3.write('abc')

    modified_file, is_dir = blocker.args
    assert modified_file in str(file3)

    # Test events in hidden folders are not emitted
    with qtbot.assertNotEmitted(fs_handler.sig_file_created, wait=2000):
        git_file = git_folder.join('git_file.txt')
        git_file.write("Some data")

    # Test events in pycache folders are not emitted
    with qtbot.assertNotEmitted(fs_handler.sig_file_created, wait=2000):
        pycache_folder = project_root.mkdir('__pycache__')
        pycache_file = pycache_folder.join("foo.pyc")
        pycache_file.write("")

    # Test events for files with binary extensions are not emitted
    with qtbot.assertNotEmitted(fs_handler.sig_file_created, wait=2000):
        png_file = project_root.join("binary.png")
        png_file.write("")


def test_loaded_and_closed_signals(create_projects, tmpdir, mocker, qtbot):
    """
    Test that loaded and closed signals are emitted when switching
    projects.
    """
    dir_object1 = tmpdir.mkdir('project1')
    path1 = to_text_string(dir_object1)
    path2 = to_text_string(tmpdir.mkdir('project2'))

    mocker.patch.object(ProjectDialog, "exec_", return_value=True)

    # Needed to actually create the files
    opened_files = []
    for file in ['file1', 'file2', 'file3']:
        file_object = dir_object1.join(file)
        file_object.write(file)
        opened_files.append(to_text_string(file_object))

    # Create the projects plugin.
    projects = create_projects(path1, opened_files)
    projects._project_types = {}
    # Switch to another project.
    with qtbot.waitSignals(
        [projects.sig_project_loaded, projects.sig_project_closed]
    ):
        projects.open_project(path=path2)


@pytest.mark.use_cli_project
def test_project_cli(projects):
    """Test that we can open a project from the command line."""
    # Simulate opening a project when the main window is visible
    projects.on_mainwindow_visible()

    # Verify that we created the expected project
    active_project = projects.get_active_project_path()
    assert osp.split(active_project)[-1] == 'cli_project_dir'

    # Close project
    projects.close_project()


def test_reopen_project(projects, tmpdir):
    """Test that we can reopen a project from the last session."""
    # Create project
    last_project = tmpdir.mkdir('last_project_dir')
    last_project.mkdir('.spyproject')
    projects.set_conf('current_project_path', str(last_project))

    # Simulate opening a project when the main window is visible
    projects.on_mainwindow_visible()

    # Verify that we created the expected project
    active_project = projects.get_active_project_path()
    assert osp.split(active_project)[-1] == 'last_project_dir'

    # Close project
    projects.close_project()


def test_recreate_project_config(projects, tmpdir):
    """
    Test that the project's config files are recreated when there are errors
    reading them.

    Regression test for spyder-ide/spyder#17907.
    """
    # Create a new directory
    path = str(tmpdir.mkdir('error_reading_config'))

    # Open project in path to generate its config
    projects.open_project(path=path)

    # Get project's config path
    config_path = projects.get_widget().current_active_project.config._path

    # Close project
    projects.close_project()

    # Append the contents of a config file to it in order to give an error
    # while reading it
    config_file = osp.join(config_path, 'workspace.ini')

    with open(config_file, 'r') as f:
        file_contents = f.readlines()

    with open(config_file, 'a') as f:
        for line in file_contents:
            f.write(line)

    # Try to read config and check we get an error
    with pytest.raises(configparser.Error):
        BaseProjectType.create_config(config_path)

    # Reopen the project again and check we recreated the config file we
    # changed
    projects.open_project(path=path)
    projects.close_project()

    with open(config_file, 'r') as f:
        new_file_contents = f.readlines()

    assert file_contents == new_file_contents


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#
"""Tests for autosave.py"""

# Standard library imports
import ast
import os.path as osp

# Third party imports
import pytest

# Local imports
from spyder.plugins.editor.utils.autosave import (AutosaveForStack,
                                                  AutosaveForPlugin)


def test_autosave_component_set_interval(mocker):
    """Test that setting the interval does indeed change it and calls
    do_autosave if enabled."""
    mocker.patch.object(AutosaveForPlugin, 'do_autosave')
    addon = AutosaveForPlugin(None)
    addon.do_autosave.assert_not_called()
    addon.interval = 10000
    assert addon.interval == 10000
    addon.do_autosave.assert_not_called()
    addon.enabled = True
    addon.interval = 20000
    assert addon.do_autosave.called


@pytest.mark.parametrize('enabled', [False, True])
def test_autosave_component_timer_if_enabled(qtbot, mocker, enabled):
    """Test that AutosaveForPlugin calls do_autosave() on timer if enabled."""
    mocker.patch.object(AutosaveForPlugin, 'do_autosave')
    addon = AutosaveForPlugin(None)
    addon.do_autosave.assert_not_called()
    addon.interval = 100
    addon.enabled = enabled
    qtbot.wait(500)
    if enabled:
        assert addon.do_autosave.called
    else:
        addon.do_autosave.assert_not_called()


def test_get_files_to_recover_with_empty_autosave_dir(mocker, tmpdir):
    """Test get_files_to_recover() when autosave dir contains no files."""
    mocker.patch('spyder.plugins.editor.utils.autosave.get_conf_path',
                 return_value=str(tmpdir))
    addon = AutosaveForPlugin(None)

    result = addon.get_files_to_recover()

    assert result == ([], [])


@pytest.mark.parametrize('running,empty',
                         [(True, False), (False, False), (False, True)])
def test_get_files_to_recover_with_one_pid_file(mocker, tmpdir,
                                                running, empty):
    """Test get_files_to_recover() if autosave dir contains one pid file with
    one autosave file. If running is True, then pretend that the pid file
    belongs to a running Spyder instance. If empty is True, then the pid file
    is empty (regression test for spyder-ide/spyder#11375)."""
    mocker.patch('spyder.plugins.editor.utils.autosave.get_conf_path',
                 return_value=str(tmpdir))
    mock_is_spyder_process = mocker.patch(
            'spyder.plugins.editor.utils.autosave.is_spyder_process',
            return_value=running)
    pidfile = tmpdir.join('pid42.txt')
    autosavefile = tmpdir.join('foo.py')
    if empty:
        pidfile.write('')
    else:
        pidfile.write('{"original": ' + repr(str(autosavefile)) + '}')
    autosavefile.write('bar = 1')
    addon = AutosaveForPlugin(None)

    result = addon.get_files_to_recover()

    if empty:  # pid file corrupted so original file name not recorded
        expected_files = [(None, str(autosavefile))]
    elif running:  # autosave file belongs to running instance
        expected_files = []
    else:
        expected_files = [('original', str(autosavefile))]
    expected = (expected_files, [str(pidfile)])
    assert result == expected
    mock_is_spyder_process.assert_called_with(42)


def test_get_files_to_recover_with_non_pid_file(mocker, tmpdir):
    """Test get_files_to_recover() if autosave dir contains no pid file, but
    one Python file."""
    mocker.patch('spyder.plugins.editor.utils.autosave.get_conf_path',
                 return_value=str(tmpdir))
    pythonfile = tmpdir.join('foo.py')
    pythonfile.write('bar = 1')
    addon = AutosaveForPlugin(None)

    result = addon.get_files_to_recover()

    expected = ([(None, str(pythonfile))], [])
    assert result == expected


def test_get_files_to_recover_without_autosave_dir(mocker):
    """Test that get_files_to_recover() does not break if there is no autosave
    directory."""
    mocker.patch('spyder.plugins.editor.utils.autosave.get_conf_path',
                 return_value='non-existing-directory')
    addon = AutosaveForPlugin(None)

    result = addon.get_files_to_recover()

    assert result == ([], [])


@pytest.mark.parametrize('error_on_remove', [False, True])
def test_try_recover(mocker, tmpdir, error_on_remove):
    """Test that try_recover_from_autosave() displays a RecoveryDialog, that
    it stores the files that the user wants to open as reported by the dialog,
    and that it removes the pid file. If error_on_remove is set, then
    removing the pid file will raise an OSError; this should be ignored."""
    mock_RecoveryDialog = mocker.patch(
            'spyder.plugins.editor.utils.autosave.RecoveryDialog')
    mocker.patch('spyder.plugins.editor.utils.autosave.get_conf_path',
                 return_value=str(tmpdir))
    pidfile = tmpdir.join('pid42.txt')
    autosavefile = tmpdir.join('foo.py')
    pidfile.write('{"original": ' + repr(str(autosavefile)) + '}')
    autosavefile.write('bar = 1')
    addon = AutosaveForPlugin(None)
    if error_on_remove:
        mocker.patch('os.remove', side_effect=OSError)

    addon.try_recover_from_autosave()

    expected_mapping = [('original', str(autosavefile))]
    mock_RecoveryDialog.assert_called_with(expected_mapping, parent=None)
    expected_files_to_open = mock_RecoveryDialog().files_to_open[:]
    assert addon.recover_files_to_open == expected_files_to_open
    if not error_on_remove:
        assert not pidfile.check()


@pytest.mark.parametrize('in_mapping,on_disk',
                         [(False, False), (True, False), (False, True)])
def test_create_unique_autosave_filename(mocker, in_mapping, on_disk):
    """Test that AutosaveForStack.create_unique_autosave_filename() returns
    a file name in the autosave directory with the same base name as the
    original file name, unless that already exists in the autosave mapping
    or on disk."""
    def new_exists(path):
        if path == osp.join('autosave', 'ham.py'):
            return on_disk
        else:
            return False

    mocker.patch('os.path.exists', side_effect=new_exists)
    addon = AutosaveForStack(mocker.Mock())
    if in_mapping:
        addon.name_mapping = {osp.join('somedir', 'ham.py'):
                              osp.join('autosave', 'ham.py')}

    autosave_filename = addon.create_unique_autosave_filename(
        osp.join('orig', 'ham.py'), 'autosave')

    if in_mapping or on_disk:
        assert autosave_filename == osp.join('autosave', 'ham-1.py')
    else:
        assert autosave_filename == osp.join('autosave', 'ham.py')


@pytest.mark.parametrize('have_hash', [True, False])
def test_autosave(mocker, have_hash):
    """Test that AutosaveForStack.maybe_autosave writes the contents to the
    autosave file and updates the file_hashes."""
    mock_editor = mocker.Mock()
    mock_fileinfo = mocker.Mock(editor=mock_editor, filename='orig',
                                newly_created=False)
    mock_document = mocker.Mock()
    mock_fileinfo.editor.document.return_value = mock_document
    mock_stack = mocker.Mock(data=[mock_fileinfo])
    addon = AutosaveForStack(mock_stack)
    addon.name_mapping = {'orig': 'autosave'}
    addon.file_hashes = {'autosave': 2}
    if have_hash:
        addon.file_hashes['orig'] = 1
    mock_stack.compute_hash.return_value = 3

    addon.maybe_autosave(0)

    mock_stack._write_to_file.assert_called_with(mock_fileinfo, 'autosave')
    mock_stack.compute_hash.assert_called_with(mock_fileinfo)
    if have_hash:
        assert addon.file_hashes == {'orig': 1, 'autosave': 3}
    else:
        assert addon.file_hashes == {'autosave': 3}


@pytest.mark.parametrize('latin', [True, False])
def test_save_autosave_mapping_with_nonempty_mapping(mocker, tmpdir, latin):
    """Test that save_autosave_mapping() writes the current autosave mapping
    to the correct file if the mapping is not empty."""
    mocker.patch('os.getpid', return_value=42)
    mocker.patch('spyder.plugins.editor.utils.autosave.get_conf_path',
                 return_value=str(tmpdir))
    addon = AutosaveForStack(None)
    if latin:
        addon.name_mapping = {'orig': 'autosave'}
    else:
        addon.name_mapping = {'原件': 'autosave'}

    addon.save_autosave_mapping()

    pidfile = tmpdir.join('pid42.txt')
    assert ast.literal_eval(pidfile.read()) == addon.name_mapping


@pytest.mark.parametrize('pidfile_exists', [False, True])
def test_save_autosave_mapping_with_empty_mapping(mocker, tmpdir,
                                                  pidfile_exists):
    """Test that save_autosave_mapping() does not write the pidfile if the
    mapping is empty, and that is removes the pidfile if it exists."""
    mocker.patch('os.getpid', return_value=42)
    mocker.patch('spyder.plugins.editor.utils.autosave.get_conf_path',
                 return_value=str(tmpdir))
    addon = AutosaveForStack(None)
    addon.name_mapping = {}
    pidfile = tmpdir.join('pid42.txt')
    if pidfile_exists:
        pidfile.write('This is an ex-parrot!')

    addon.save_autosave_mapping()

    assert not pidfile.check()


@pytest.mark.parametrize('exception', [False, True])
def test_autosave_remove_autosave_file(mocker, exception):
    """Test that AutosaveForStack.remove_autosave_file removes the autosave
    file, that an error dialog is displayed if an exception is raised,
    and that the autosave file is removed from `name_mapping` and
    `file_hashes`."""
    mock_remove = mocker.patch('os.remove')
    if exception:
        mock_remove.side_effect = OSError()
    mock_dialog = mocker.patch(
        'spyder.plugins.editor.utils.autosave.AutosaveErrorDialog')
    mock_stack = mocker.Mock()
    fileinfo = mocker.Mock()
    fileinfo.filename = 'orig'
    addon = AutosaveForStack(mock_stack)
    addon.name_mapping = {'orig': 'autosave'}
    addon.file_hashes = {'autosave': 42}

    addon.remove_autosave_file(fileinfo.filename)

    assert addon.name_mapping == {}
    assert addon.file_hashes == {}
    mock_remove.assert_any_call('autosave')
    assert mock_dialog.called == exception


def test_get_autosave_filename(mocker, tmpdir):
    """Test that AutosaveForStack.get_autosave_filename returns a consistent
    and unique name for the autosave file is returned."""
    addon = AutosaveForStack(mocker.Mock())
    mocker.patch('spyder.plugins.editor.utils.autosave.get_conf_path',
                 return_value=str(tmpdir))

    expected = str(tmpdir.join('foo.py'))
    assert addon.get_autosave_filename('foo.py') == expected

    expected2 = str(tmpdir.join('foo-1.py'))
    assert addon.get_autosave_filename('foo.py') == expected
    assert addon.get_autosave_filename('ham/foo.py') == expected2


@pytest.mark.parametrize('have_hash', [True, False])
def test_autosave_file_renamed(mocker, tmpdir, have_hash):
    """Test that AutosaveForStack.file_renamed removes the old autosave file,
    creates a new one, and updates `name_mapping` and `file_hashes`."""
    mock_remove = mocker.patch('os.remove')
    mocker.patch('spyder.plugins.editor.utils.autosave.get_conf_path',
                 return_value=str(tmpdir))
    mock_editor = mocker.Mock()
    mock_fileinfo = mocker.Mock(editor=mock_editor, filename='new_foo.py',
                                newly_created=False)
    mock_document = mocker.Mock()
    mock_fileinfo.editor.document.return_value = mock_document
    mock_stack = mocker.Mock(data=[mock_fileinfo])
    mock_stack.has_filename.return_value = 0
    mock_stack.compute_hash.return_value = 3
    addon = AutosaveForStack(mock_stack)
    old_autosavefile = str(tmpdir.join('old_foo.py'))
    new_autosavefile = str(tmpdir.join('new_foo.py'))
    addon.name_mapping = {'old_foo.py': old_autosavefile}
    addon.file_hashes = {'old_foo.py': 1, old_autosavefile: 42}
    if have_hash:
        addon.file_hashes = {'old_foo.py': 1, old_autosavefile: 42}
    else:
        addon.file_hashes = {old_autosavefile: 42}

    addon.file_renamed('old_foo.py', 'new_foo.py')

    mock_remove.assert_any_call(old_autosavefile)
    mock_stack._write_to_file.assert_called_with(
        mock_fileinfo, new_autosavefile)
    assert addon.name_mapping == {'new_foo.py': new_autosavefile}
    if have_hash:
        assert addon.file_hashes == {'new_foo.py': 1, new_autosavefile: 3}
    else:
        assert addon.file_hashes == {new_autosavefile: 3}


if __name__ == "__main__":
    pytest.main()
